from typing import Any, Dict
from ms_deisotope.utils import dict_proxy
from .cv import Term, TermSet


class SoftwareName(Term):
    pass


@dict_proxy('options')
class Software(object):

    name: str
    id: str
    version: str
    options: Dict[str, Any]

    @classmethod
    def is_name(cls, name):
        return name in software_names_by_name

    def __init__(self, name=None, id=None, version=None, **options):
        if name is None:
            name, options = self._resolve_name_from_kwargs(options)
        if name is None and id is not None:
            name = id
        self.name = name
        self.id = id or name
        self.version = version
        self.options = options

    def _resolve_name_from_kwargs(self, options: Dict[str, Any]):
        names = dict()
        not_names = dict()
        for key, value in options.items():
            if self.is_name(key):
                names[key] = value
            else:
                not_names[key] = value
        options = not_names
        if len(names) == 1:
            name = list(names.keys())[0]
        elif len(names) == 0:
            name = None
        else:
            raise ValueError("Multiple possible names found")
        return name, options

    def __str__(self):
        return self.name

    def __repr__(self):
        template = '{self.__class__.__name__}({self.name}, {self.id}, {self.version})'
        return template.format(self=self)

    def __eq__(self, other):
        try:
            return self.name == other.name
        except AttributeError:
            return str(self) == str(other)


software_names = []

# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('software', list_name='software_names', term_cls_name="SoftwareName", writer=cog.out)
# ]]]
# CV Version: 4.1.95
software_names = TermSet([
    SoftwareName('Agilent software', 'MS:1000689',
                 ('Agilent software for data acquisition and analysis.'),
                 'software',
                 ['software']),
    SoftwareName('SCIEX software', 'MS:1000690',
                 ('SCIEX or Applied Biosystems software for data acquisition '
                  'and analysis.'),
                 'software',
                 ['software']),
    SoftwareName('Applied Biosystems software', 'MS:1000691',
                 ('Applied Biosystems|MDS SCIEX software for data acquisition '
                  'and analysis.'),
                 'software',
                 ['software']),
    SoftwareName('Bruker software', 'MS:1000692',
                 ('Bruker software for data acquisition and analysis.'),
                 'software',
                 ['software']),
    SoftwareName('Thermo Finnigan software', 'MS:1000693',
                 ('Thermo Finnigan software for data acquisition and analysis.'),
                 'software',
                 ['software']),
    SoftwareName('Waters software', 'MS:1000694',
                 ('Waters software for data acquisition and analysis.'),
                 'software',
                 ['software']),
    SoftwareName('custom unreleased software tool', 'MS:1000799',
                 ('A software tool that has not yet been released. The value '
                  'should describe the software. Please do not use this term '
                  'for publicly available software - contact the PSI-MS working '
                  'group in order to have another CV term added.'),
                 'software',
                 ['software']),
    SoftwareName('SRM software', 'MS:1000871',
                 ('Software used to predict, select, or optimize transitions or '
                  'analyze the results of selected reaction monitoring runs.'),
                 'software',
                 ['software']),
    SoftwareName('peptide attribute calculation software', 'MS:1000873',
                 ('Software used to predict or calculate numerical attributes '
                  'of peptides.'),
                 'software',
                 ['software']),
    SoftwareName('quantitation software name', 'MS:1001139',
                 ('Quantitation software name.'),
                 'software',
                 ['software', 'quantification information']),
    SoftwareName('acquisition software', 'MS:1001455',
                 ('Acquisition software.'),
                 'software',
                 ['software']),
    SoftwareName('analysis software', 'MS:1001456',
                 ('Analysis software.'),
                 'software',
                 ['software']),
    SoftwareName('data processing software', 'MS:1001457',
                 ('Data processing software.'),
                 'software',
                 ['software']),
    SoftwareName('Shimadzu Corporation software', 'MS:1001557',
                 ('Shimadzu Corporation software.'),
                 'software',
                 ['software']),
    SoftwareName('LECO software', 'MS:1001798',
                 ('LECO software for data acquisition and analysis.'),
                 'software',
                 ['software']),
    SoftwareName('BSI software', 'MS:1001949',
                 ('Bioinformatics Solutions Inc. Software for data processing '
                  'and analysis.'),
                 'software',
                 ['software']),
    SoftwareName('SCiLS software', 'MS:1002383',
                 ('SCiLS software for data acquisition and analysis.'),
                 'software',
                 ['software']),
    SoftwareName('library creation software', 'MS:1003207',
                 ('Library creation software'),
                 'software',
                 ['software', 'spectral library attribute']),
    SoftwareName('MassHunter Data Acquisition', 'MS:1000678',
                 ('Software for data acquisition of 6000 series instruments.'),
                 'software',
                 ['Agilent software', 'acquisition software', 'software']),
    SoftwareName('MassHunter Easy Access', 'MS:1000679',
                 ('Software for open access data acquisition.'),
                 'software',
                 ['Agilent software', 'acquisition software', 'software']),
    SoftwareName('MassHunter Qualitative Analysis', 'MS:1000680',
                 ('Software for data analysis of data from 6000 series '
                  'instruments.'),
                 'software',
                 ['Agilent software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MassHunter Quantitative Analysis', 'MS:1000681',
                 ('Software for quantitation of Triple Quadrupole and '
                  'Quadrupole Time-of-Flight data.'),
                 'software',
                 ['Agilent software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MassHunter Metabolite ID', 'MS:1000682',
                 ('Software for identification of metabolites.'),
                 'software',
                 ['Agilent software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MassHunter BioConfirm', 'MS:1000683',
                 ('Software for protein characterization.'),
                 'software',
                 ['Agilent software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Genespring MS', 'MS:1000684',
                 ('Software for quantitation and statistical analysis of TOF '
                  'and Q-TOF LC/MS data.'),
                 'software',
                 ['Agilent software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MassHunter Mass Profiler', 'MS:1000685',
                 ('Software for quantitation and statistical analysis of TOF '
                  'and Q-TOF LC/MS data.'),
                 'software',
                 ['Agilent software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('METLIN', 'MS:1000686',
                 ('Personal Metabolite Database for MassHunter Workstation. '
                  'Software for identification of human metabolites.'),
                 'software',
                 ['Agilent software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Spectrum Mill for MassHunter Workstation', 'MS:1000687',
                 ('Software for protein identification and characterization of '
                  'complex protein digest mixtures.'),
                 'software',
                 ['Agilent software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('6300 Series Ion Trap Data Analysis Software', 'MS:1000688',
                 ('Software for data analysis of 6300 series ion trap mass '
                  'spectrometers.'),
                 'software',
                 ['Agilent software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Analyst', 'MS:1000551',
                 ('SCIEX or Applied Biosystems|MDS SCIEX software for data '
                  'acquisition.'),
                 'software',
                 ['SCIEX software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('4000 Series Explorer Software', 'MS:1000659',
                 ('SCIEX or Applied Biosystems software for data acquisition '
                  'and analysis.'),
                 'software',
                 ['SCIEX software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('GPS Explorer', 'MS:1000661',
                 ('SCIEX or Applied Biosystems software for data acquisition '
                  'and analysis.'),
                 'software',
                 ['SCIEX software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('LightSight Software', 'MS:1000662',
                 ('SCIEX or Applied Biosystems|MDS SCIEX software metabolite '
                  'identification.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('ProteinPilot Software', 'MS:1000663',
                 ('SCIEX or Applied Biosystems|MDS SCIEX software for protein '
                  'ID and quant.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TissueView Software', 'MS:1000664',
                 ('Applied Biosystems|MDS SCIEX software for tissue imaging.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MarkerView Software', 'MS:1000665',
                 ('Applied Biosystems|MDS SCIEX software for metabolomics and '
                  'biomarker profiling.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MRMPilot Software', 'MS:1000666',
                 ('Applied Biosystems|MDS SCIEX software for MRM assay '
                  'development.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('BioAnalyst', 'MS:1000667',
                 ('Applied Biosystems|MDS SCIEX software for bio-related data '
                  'exploration.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Pro ID', 'MS:1000668',
                 ('Applied Biosystems|MDS SCIEX software for protein '
                  'identification.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Pro ICAT', 'MS:1000669',
                 ('Applied Biosystems|MDS SCIEX software for protein ID and '
                  'quant by ICAT.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Pro Quant', 'MS:1000670',
                 ('Applied Biosystems|MDS SCIEX software for protein ID and '
                  'quant by iTRAQ.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Pro BLAST', 'MS:1000671',
                 ('Applied Biosystems|MDS SCIEX software for MS-BLAST '
                  'identification.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Cliquid', 'MS:1000672',
                 ('SCIEX Cliquid software for data analysis and quantitation.'),
                 'software',
                 ['SCIEX software', 'software']),
    SoftwareName('MIDAS Workflow Designer', 'MS:1000673',
                 ('Applied Biosystems|MDS SCIEX software for MRM assay '
                  'development.'),
                 'software',
                 ['SCIEX software', 'software']),
    SoftwareName('MultiQuant', 'MS:1000674',
                 ('Applied Biosystems|MDS SCIEX software for MRM-based '
                  'quantitation.'),
                 'software',
                 ['SCIEX software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('SCIEX TOF/TOF Series Explorer Software', 'MS:1001483',
                 ('SCIEX or Applied Biosystems software for TOF/TOF data '
                  'acquisition and analysis.'),
                 'software',
                 ['SCIEX software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Data Explorer', 'MS:1000536',
                 ('Applied Biosystems software for data acquisition and '
                  'analysis.'),
                 'software',
                 ['Applied Biosystems software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('4700 Explorer', 'MS:1000537',
                 ('Applied Biosystems software for data acquisition and '
                  'analysis.'),
                 'software',
                 ['Applied Biosystems software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Voyager Biospectrometry Workstation System', 'MS:1000539',
                 ('Applied Biosystems MALDI-TOF data acquisition and analysis '
                  'system.'),
                 'software',
                 ['Applied Biosystems software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('FlexAnalysis', 'MS:1000535',
                 ('Bruker software for data analysis.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('FlexControl', 'MS:1000540',
                 ('Bruker software for data acquisition.'),
                 'software',
                 ['Bruker software', 'acquisition software', 'software']),
    SoftwareName('apexControl', 'MS:1000706',
                 ('Bruker software for data acquisition.'),
                 'software',
                 ['Bruker software', 'acquisition software', 'software']),
    SoftwareName('BioTools', 'MS:1000707',
                 ('Bruker software for data analysis.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('CLINPROT', 'MS:1000708',
                 ('Bruker CLINPROT software.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('CLINPROT micro', 'MS:1000709',
                 ('Bruker CLINPROT micro software.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('CLINPROT robot', 'MS:1000710',
                 ('Bruker CLINPROT robot software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('ClinProTools', 'MS:1000711',
                 ('Bruker ClinProTools software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('Compass', 'MS:1000712',
                 ('Bruker Compass software.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Compass for HCT/esquire', 'MS:1000713',
                 ('Bruker Compass for HCT/esquire software.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Compass for micrOTOF', 'MS:1000714',
                 ('Bruker Compass for micrOTOF software.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Compass OpenAccess', 'MS:1000715',
                 ('Bruker compass OpenAccess software.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Compass Security Pack', 'MS:1000716',
                 ('Bruker compass Security Pack software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('CompassXport', 'MS:1000717',
                 ('Bruker stand-alone software for data conversion.'),
                 'software',
                 ['Bruker software', 'data processing software', 'software']),
    SoftwareName('CompassXtract', 'MS:1000718',
                 ('Bruker software library for data access.'),
                 'software',
                 ['Bruker software', 'data processing software', 'software']),
    SoftwareName('DataAnalysis', 'MS:1000719',
                 ('Bruker software for data analysis.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('dpControl', 'MS:1000720',
                 ('Bruker software for data acquisition.'),
                 'software',
                 ['Bruker software', 'acquisition software', 'software']),
    SoftwareName('esquireControl', 'MS:1000721',
                 ('Bruker software for data acquisition.'),
                 'software',
                 ['Bruker software', 'acquisition software', 'software']),
    SoftwareName('flexImaging', 'MS:1000722',
                 ('Bruker software for data analysis.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('GENOLINK', 'MS:1000723',
                 ('Bruker GENOLINK software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('GenoTools', 'MS:1000724',
                 ('Bruker GenoTools software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('HCTcontrol', 'MS:1000725',
                 ('Bruker software for data acquisition.'),
                 'software',
                 ['Bruker software', 'acquisition software', 'software']),
    SoftwareName('micrOTOFcontrol', 'MS:1000726',
                 ('Bruker software for data acquisition.'),
                 'software',
                 ['Bruker software', 'acquisition software', 'software']),
    SoftwareName('PolyTools', 'MS:1000727',
                 ('Bruker PolyTools software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('ProfileAnalysis', 'MS:1000728',
                 ('Bruker software for data analysis.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('PROTEINEER', 'MS:1000729',
                 ('Bruker PROTEINEER software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('PROTEINEER dp', 'MS:1000730',
                 ('Bruker PROTEINEER dp software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('PROTEINEER fc', 'MS:1000731',
                 ('Bruker PROTEINEER fc software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('PROTEINEER spII', 'MS:1000732',
                 ('Bruker PROTEINEER spII software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('PROTEINEER-LC', 'MS:1000733',
                 ('Bruker PROTEINEER-LC software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('ProteinScape', 'MS:1000734',
                 ('Bruker ProteinScape software.'),
                 'software',
                 ['Bruker software', 'analysis software', 'software']),
    SoftwareName('PureDisk', 'MS:1000735',
                 ('BrukerPureDisk software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('QuantAnalysis', 'MS:1000736',
                 ('Bruker software for data analysis.'),
                 'software',
                 ['Bruker software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('spControl', 'MS:1000737',
                 ('Bruker software for data acquisition.'),
                 'software',
                 ['Bruker software', 'acquisition software', 'software']),
    SoftwareName('TargetAnalysis', 'MS:1000738',
                 ('Bruker TargetAnalysis software.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('WARP-LC', 'MS:1000739',
                 ('Bruker WARP-LC software.'),
                 'software',
                 ['Bruker software', 'quantitation software name', 'software', 'quantification information']),
    SoftwareName('HyStar', 'MS:1000817',
                 ('Bruker software for hyphenated experiments.'),
                 'software',
                 ['Bruker software', 'software']),
    SoftwareName('ProteinExtractor', 'MS:1001487',
                 ('An algorithm for protein determination/assembly integrated '
                  "into Bruker's ProteinScape."),
                 'software',
                 ['Bruker software', 'analysis software', 'software']),
    SoftwareName('Xcalibur', 'MS:1000532',
                 ('Thermo Finnigan software for data acquisition and analysis.'),
                 'software',
                 ['Thermo Finnigan software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Bioworks', 'MS:1000533',
                 ('Thermo Finnigan software for data analysis of peptides and '
                  'proteins.'),
                 'software',
                 ['Thermo Finnigan software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Proteome Discoverer', 'MS:1000650',
                 ('Thermo Scientific software for data analysis of peptides and '
                  'proteins.'),
                 'software',
                 ['Thermo Finnigan software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('PinPoint', 'MS:1001912',
                 ('Thermo Scientific PinPoint SRM analysis software.'),
                 'software',
                 ['Thermo Finnigan software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MassLynx', 'MS:1000534',
                 ('Micromass software for data acquisition and analysis.'),
                 'software',
                 ['Waters software', 'acquisition software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('ProteinLynx Global Server', 'MS:1000601',
                 ('Waters software for data analysis.'),
                 'software',
                 ['Waters software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Empower', 'MS:1001795',
                 ('Waters Empower software for liquid chromatography and mass '
                  'spectrometry acquisition.'),
                 'software',
                 ['Waters software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('UNIFY', 'MS:1001796',
                 ('Waters UNIFY software for liquid chromatography and mass '
                  'spectrometry acquisition.'),
                 'software',
                 ['Waters software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MaRiMba', 'MS:1000872',
                 ('Software used to predict transitions for selected reaction '
                  'monitoring experiments based on observed spectrum libraries '
                  'developed and distributed by the Institute for Systems '
                  'Biology.'),
                 'software',
                 ['SRM software', 'software']),
    SoftwareName('Skyline', 'MS:1000922',
                 ('Software used to predict, select, and optimize transitions '
                  'as well as analyze the results of selected reaction '
                  'monitoring runs developed and distributed by the MacCoss lab '
                  'at the University of Washington.'),
                 'software',
                 ['SRM software', 'quantitation software name', 'software', 'quantification information']),
    SoftwareName('TIQAM', 'MS:1000923',
                 ('Software used to predict, select, and optimize transitions '
                  'for selected reaction monitoring experiments developed and '
                  'distributed by the Institute for Systems Biology.'),
                 'software',
                 ['SRM software', 'software']),
    SoftwareName('ATAQS', 'MS:1000925',
                 ('Software suite used to predict, select, and optimize '
                  'transitions as well as analyze the results of selected '
                  'reaction monitoring runs developed and distributed by the '
                  'Institute for Systems Biology.'),
                 'software',
                 ['SRM software', 'software']),
    SoftwareName('MRMaid', 'MS:1002220',
                 ('A web-based SRM assay design tool whose transitions are '
                  'generated by mining the millions of identified peptide '
                  "spectra held in the EBI's PRIDE database."),
                 'software',
                 ['SRM software', 'software']),
    SoftwareName('Anubis', 'MS:1002410',
                 ('Anubis software for selected reaction monitoring data." '
                  '[PSI:PI'),
                 'software',
                 ['SRM software', 'quantitation software name', 'software', 'quantification information']),
    SoftwareName('SSRCalc', 'MS:1000874',
                 ('Sequence Specific Retention Calculator estimates the '
                  'retention time of peptides based on their sequence.'),
                 'software',
                 ['peptide attribute calculation software', 'software']),
    SoftwareName('Mascot Distiller', 'MS:1001488',
                 ('Mascot Distiller.'),
                 'software',
                 ['quantitation software name', 'analysis software', 'software', 'quantification information']),
    SoftwareName('MaxQuant', 'MS:1001583',
                 ('MaxQuant is a quantitative proteomics software package '
                  'designed for analyzing large mass spectrometric data sets. '
                  'It is specifically aimed at high resolution MS data.'),
                 'software',
                 ['quantitation software name', 'analysis software', 'software', 'quantification information']),
    SoftwareName('Progenesis LC-MS', 'MS:1001830',
                 ('Software from Nonlinear Dynamics for LC-MS label-free '
                  'workflow.'),
                 'software',
                 ['quantitation software name', 'software', 'quantification information']),
    SoftwareName('SILACAnalyzer', 'MS:1001831',
                 ('Software for SILAC workflow.'),
                 'software',
                 ['quantitation software name', 'TOPP software', 'software', 'quantification information', 'analysis software', 'data processing software']),
    SoftwareName('PEAKS Studio', 'MS:1001946',
                 ('PEAKS Studio software for data analysis.'),
                 'software',
                 ['quantitation software name', 'analysis software', 'data processing software', 'software', 'quantification information']),
    SoftwareName('PEAKS Online', 'MS:1001947',
                 ('PEAKS Online software for high throughput data analysis.'),
                 'software',
                 ['quantitation software name', 'analysis software', 'data processing software', 'software', 'quantification information']),
    SoftwareName('PEAKS Node', 'MS:1001948',
                 ('PEAKS Node software for high throughput data analysis.'),
                 'software',
                 ['quantitation software name', 'analysis software', 'data processing software', 'software', 'quantification information']),
    SoftwareName('Microsoft Excel', 'MS:1002059',
                 ('Microsoft Excel (can be used for spectral counting).'),
                 'software',
                 ['quantitation software name', 'software', 'quantification information']),
    SoftwareName('FindPairs', 'MS:1002063',
                 ('Software e.g. for SILAC and 14N/15N workflow, part of the '
                  'PeakQuant suite.'),
                 'software',
                 ['quantitation software name', 'software', 'quantification information']),
    SoftwareName('x-Tracker', 'MS:1002123',
                 ('X-Tracker generic tool for quantitative proteomics.'),
                 'software',
                 ['quantitation software name', 'software', 'quantification information']),
    SoftwareName('ProteoSuite', 'MS:1002124',
                 ('ProteoSuite software for the analysis of quantitative '
                  'proteomics data." [DOI:10.1089/omi.2012.0022, PMID:22804616'),
                 'software',
                 ['quantitation software name', 'software', 'quantification information']),
    SoftwareName('ITRAQAnalyzer', 'MS:1002129',
                 ('Software for iTRAQ workflow. Extracts and normalizes iTRAQ '
                  'information from an MS experiment.'),
                 'software',
                 ['quantitation software name', 'TOPP software', 'software', 'quantification information', 'analysis software', 'data processing software']),
    SoftwareName('IsobariQ', 'MS:1002210',
                 ('A quantitative software package designed for analysis of '
                  'IPTL, TMT and iTRAQ data." [PMID:21067241, '
                  'DOI:10.1021/pr1009977'),
                 'software',
                 ['quantitation software name', 'analysis software', 'software', 'quantification information']),
    SoftwareName('Skyline mzQuantML converter', 'MS:1002546',
                 ('A software package to convert Skyline report to mzQuantML." '
                  '[PSI:PI'),
                 'software',
                 ['quantitation software name', 'analysis software', 'software', 'quantification information']),
    SoftwareName('PatternLab', 'MS:1003108',
                 ('PatternLab for Proteomics is an integrated computational '
                  'environment for analyzing shotgun proteomic data.'),
                 'software',
                 ['quantitation software name', 'analysis software', 'software', 'quantification information']),
    SoftwareName('QUIN-XL', 'MS:1003111',
                 ('Quantification of cross-linked peptides in complex protein '
                  'mixtures'),
                 'software',
                 ['quantitation software name', 'software', 'quantification information']),
    SoftwareName('DIA-NN', 'MS:1003253',
                 ('A universal software for data-independent acquisition (DIA) '
                  'proteomics data processing" [PMID:31768060'),
                 'software',
                 ['quantitation software name', 'analysis software', 'software', 'quantification information']),
    SoftwareName('MALDI Solutions', 'MS:1001558',
                 ('Shimadzu Biotech software for data acquisition, processing, '
                  'and analysis.'),
                 'software',
                 ['acquisition software', 'analysis software', 'data processing software', 'Shimadzu Corporation software', 'software']),
    SoftwareName('ChromaTOF software', 'MS:1001799',
                 ('Software for acquisition, processing and analysis of data '
                  'for LECO instruments.'),
                 'software',
                 ['acquisition software', 'analysis software', 'data processing software', 'LECO software', 'software']),
    SoftwareName('ChromaTOF HRT software', 'MS:1001877',
                 ('Software for acquisition, processing and analysis of data '
                  'for LECO instruments.'),
                 'software',
                 ['acquisition software', 'analysis software', 'data processing software', 'LECO software', 'software']),
    SoftwareName('MALDI Solutions LC-MALDI', 'MS:1002381',
                 ('Software for automated LC-MALDI analysis and reporting.'),
                 'software',
                 ['acquisition software', 'analysis software', 'data processing software', 'Shimadzu Corporation software', 'software']),
    SoftwareName('Proteios', 'MS:1000600',
                 ('Database application and analysis platform for proteomics." '
                  '[PSI:MS'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('ProteoWizard software', 'MS:1000615',
                 ('ProteoWizard software for data processing and analysis. '
                  'Primarily developed by the labs of P. Malick and D. Tabb.'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP software', 'MS:1000752',
                 ('TOPP (The OpenMS proteomics pipeline) software.'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('Mascot', 'MS:1001207',
                 ('The name of the Mascot search engine.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('SEQUEST', 'MS:1001208',
                 ('The name of the SEQUEST search engine.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Phenyx', 'MS:1001209',
                 ('The name of the Phenyx search engine.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('greylag', 'MS:1001461',
                 ('Greylag identification software.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('OMSSA', 'MS:1001475',
                 ('Open Mass Spectrometry Search Algorithm was used to analyze '
                  'the spectra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('X\\!Tandem', 'MS:1001476',
                 ('X!Tandem was used to analyze the spectra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('SpectraST', 'MS:1001477',
                 ('Open-source software for mass spectral library creation and '
                  'searching, developed at the Institute for Systems Biology '
                  'and the Hong Kong University of Science and Technology. Part '
                  'of the Trans-Proteomic Pipeline.'),
                 'software',
                 ['analysis software', 'library creation software', 'software', 'spectral library attribute']),
    SoftwareName('Mascot Parser', 'MS:1001478',
                 ('Mascot Parser was used to analyze the spectra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Mascot Integra', 'MS:1001489',
                 ('Mascot Integra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Percolator', 'MS:1001490',
                 ('Percolator.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Scaffold', 'MS:1001561',
                 ('Scaffold analysis software.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('XCMS', 'MS:1001582',
                 ('Bioconductor package XCMS for preprocessing high-throughput, '
                  'untargeted analyte profiling data.'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('MyriMatch', 'MS:1001585',
                 ('Tabb Lab software for directly comparing peptides in a '
                  'database to tandem mass spectra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('DirecTag', 'MS:1001586',
                 ('Tabb Lab software for generating sequence tags from tandem '
                  'mass spectra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('TagRecon', 'MS:1001587',
                 ('Tabb Lab software for reconciling sequence tags to a protein '
                  'database.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Pepitome', 'MS:1001588',
                 ('Tabb Lab software for spectral library searches on tandem '
                  'mass spectra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('SQID', 'MS:1001886',
                 ('Software for data analysis of peptides and proteins.'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('DeBunker', 'MS:1001973',
                 ('DeBunker software.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('MSQuant', 'MS:1001977',
                 ('MSQuant software.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Ascore software', 'MS:1001984',
                 ('Ascore software.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('ProteinProspector', 'MS:1002043',
                 ('ProteinProspector software for data acquisition and '
                  'analysis.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('MS-GF', 'MS:1002047',
                 ('MS-GF software used to re-score the peptide-spectrum '
                  'matches." [DOI:10.1074/mcp.M110.003731'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('MS-GF+', 'MS:1002048',
                 ('MS-GF+ software used to analyze the spectra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('PAnalyzer', 'MS:1002076',
                 ('PAnalyzer software for getting protein evidence categories.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('mzidLib', 'MS:1002237',
                 ('A library of Java routines for manipulating mzIdentML files.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Comet', 'MS:1002251',
                 ('Comet open-source sequence search engine developed at the '
                  'University of Washington.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Byonic', 'MS:1002261',
                 ('Byonic search engine from Protein Metrics.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Trans-Proteomic Pipeline', 'MS:1002285',
                 ('A suite of open source tools for the processing of MS2 '
                  'proteomics data developed by the Seattle Proteome Center at '
                  'the Institute for Systems Biology.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Trans-Proteomic Pipeline software', 'MS:1002286',
                 ('A software program that is a component of the Trans- '
                  'Proteomic Pipeline.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Amanda', 'MS:1002336',
                 ('Amanda scoring system for PSM identification.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Andromeda', 'MS:1002337',
                 ('Andromeda is a peptide search engine.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('MZmine', 'MS:1002342',
                 ('A framework for differential analysis of mass spectrometry '
                  'data." [PMID:16403790'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('Maltcms', 'MS:1002344',
                 ('Modular Application Toolkit for Chromatography Mass- '
                  'Spectrometry is an application framework mainly for '
                  'developers." [PSI:PI'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('SCiLS Lab', 'MS:1002384',
                 ('SCiLS Lab software.'),
                 'software',
                 ['SCiLS software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('PIA', 'MS:1002387',
                 ('PIA - Protein Inference Algorithms, a toolbox for protein '
                  'inference and identification analysis." [PSI:PI'),
                 'software',
                 ['postprocessing software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Maui', 'MS:1002452',
                 ('The Maltcms Graphical User Interface." [PSI:PI'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('PeptideShaker', 'MS:1002458',
                 ('PeptideShaker is a software for the interpretation of '
                  'proteomics identification results." [PSI:PI'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('xiFDR', 'MS:1002543',
                 ('Target/Decoy based FDR estimation for cross-linking peptide- '
                  'identifications.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('xi', 'MS:1002544',
                 ('Search engine for cross-linked peptides.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Tide', 'MS:1002575',
                 ('Tide open-source sequence search program developed at the '
                  'University of Washington.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('ProLuCID', 'MS:1002596',
                 ('The SEQUEST-like sequence search engine ProLuCID, developed '
                  'in the Yates Lab at the Scripps Research Institute.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('DTASelect', 'MS:1002598',
                 ('Analysis software designed to reassemble the SEQUEST peptide '
                  'identifications and to highlight the most significant '
                  'matches." [PMID:12643522'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('MSDK', 'MS:1002645',
                 ('Mass Spectrometry Development Kit (MSDK) is a Java library '
                  'of algorithms for processing of mass spectrometry data." '
                  '[PSI:PI'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('Morpheus', 'MS:1002661',
                 ('Morpheus search engine.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('MSPathFinder', 'MS:1002720',
                 ('PNNL top-down/bottom-up analysis software for identifying '
                  'peptides and proteoforms in fragmentation mass spectra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('NIST MSPepSearch', 'MS:1002750',
                 ('Search tool of the NIST (National Institute of Standards and '
                  'Technology) for spectral library searches.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('MetaMorpheus', 'MS:1002826',
                 ('MetaMorpheus search engine.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('mzR', 'MS:1002869',
                 ('Bioconductor package mzR for reading and writing mass '
                  'spectrometry data files.'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('MSnbase', 'MS:1002870',
                 ('Bioconductor package MSnbase provides infrastructure for '
                  'manipulation, processing and visualization of mass '
                  'spectrometry and proteomics data, ranging from raw to '
                  'quantitative and annotated data.'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('CAMERA', 'MS:1002871',
                 ('Bioconductor package CAMERA for annotation of peak lists '
                  'generated by xcms, rule based annotation of isotopes and '
                  'adducts, isotope validation, EIC correlation based tagging '
                  'of unknown adducts and fragments.'),
                 'software',
                 ['analysis software', 'data processing software', 'software']),
    SoftwareName('small molecule analysis software', 'MS:1002878',
                 ('Software for the analysis of small molecules.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('TopPIC', 'MS:1002901',
                 ('TopPIC: a software tool for top-down mass spectrometry-based '
                  'proteoform identification and characterization." '
                  '[PMID:27423895'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('TopFD', 'MS:1002902',
                 ('Top-down mass spectral feature detection.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('TopMG', 'MS:1002903',
                 ('A mass graph-based approach for the identification of '
                  'modified proteoforms using top-down tandem mass spectra." '
                  '[PMID:28453668'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Proline', 'MS:1002981',
                 ('The Proline software suite for mass spectrometry based '
                  'proteomics.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('PepNovo', 'MS:1002982',
                 ('PepNovo tool for de novo peptide sequencing.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('pNovo', 'MS:1002983',
                 ('pNovo tool for de novo peptide sequencing and identification '
                  'using HCD spectra.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Novor', 'MS:1002984',
                 ('Novor real-time peptide de novo sequencing software tool.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('IdentiPy', 'MS:1002987',
                 ('IdentiPy." [PMID:29682971'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('pFind', 'MS:1003011',
                 ('Sequence-tag-based search engine pFind.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('i3tms', 'MS:1003013',
                 ('i3-tms search engine and data-analysis software.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('MSFragger', 'MS:1003014',
                 ('A database search-based peptide identification tool.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('Philosopher', 'MS:1003018',
                 ('General proteomics processing toolkit for shotgun '
                  'proteomics.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('SIM-XL', 'MS:1003109',
                 ('Identifying cross-linked peptides in complex protein '
                  'mixtures'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('EPIFANY', 'MS:1003118',
                 ('A Method for Efficient High-Confidence Protein Inference. '
                  'The tool is part of the OpenMS framework'),
                 'software',
                 ['analysis software', 'TOPP software', 'software', 'data processing software']),
    SoftwareName('ProSight', 'MS:1003141',
                 ('ProSight: Database search engine for top-down proteomics.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('TDPortal', 'MS:1003142',
                 ('TDPortal: Database search engine for top-down proteomics.'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('pyteomics', 'MS:1003146',
                 ('Python module that helps handling various proteomics data '
                  'analysis tasks." [DOI:10.1021/acs.jproteome.8b00717'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('PTX-QC', 'MS:1003162',
                 ('Proteomics (PTX) - QualityControl (QC) software for QC '
                  'report generation and visualization." '
                  '[DOI:10.1021/acs.jproteome.5b00780, PMID:26653327'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('QuaMeter IDFree', 'MS:1003164',
                 ('QuaMeter IDFree software for QC metric calculation." '
                  '[DOI:10.1021/ac4034455'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('iMonDB', 'MS:1003165',
                 ('iMonDB software to extract, store, and manage mass '
                  'spectrometry instrument parameters from raw data files." '
                  '[DOI:10.1021/acs.jproteome.5b00127, PMID:25798920'),
                 'software',
                 ['analysis software', 'software']),
    SoftwareName('BiblioSpec', 'MS:1003202',
                 ('A suite of software tools for creating and searching MS/MS '
                  'peptide spectrum libraries, developed at the University of '
                  'Washington'),
                 'software',
                 ['analysis software', 'library creation software', 'software', 'spectral library attribute']),
    SoftwareName('PeakForest', 'MS:1003232',
                 ('comprehensive infrastructure to organize, curate and share a '
                  'multi- instrument spectral library for metabolomics data '
                  'annotation developed and distributed by the French National '
                  'infrastructure in metabolomics and fluxomics (MetaboHUB).'),
                 'software',
                 ['analysis software', 'library creation software', 'small molecule analysis software', 'software', 'spectral library attribute']),
    SoftwareName('massWolf', 'MS:1000538',
                 ('A software for converting Waters raw directory format to '
                  'mzXML or mzML. MassWolf was originally developed at the '
                  'Institute for Systems Biology.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('ReAdW', 'MS:1000541',
                 ('A software program for converting Thermo Finnigan RAW file '
                  'format to mzXML or mzML. ReAdW was originally developed at '
                  'the Institute for Systems Biology. Its whimsical interleaved '
                  'spelling and capitalization is pronounced \\"readraw\\".'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('MzStar', 'MS:1000542',
                 ('A software program for converting Applied Biosystems wiff '
                  'file format to mzXML format. MzStar was originally developed '
                  'at the Institute for Systems Biology. It is now obsoleted by '
                  'the MzWiff program.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('Trapper', 'MS:1000553',
                 ('A software program for converting Agilent MassHunter format '
                  'to mzXML or mzML. Trapper was originally developed at the '
                  'Institute for Systems Biology.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('MzWiff', 'MS:1000591',
                 ('A software program for converting Applied Biosystems wiff '
                  'file format to the mzXML or mzML format. MzWiff is currently '
                  'maintained at the Institute for Systems Biology. It replaces '
                  'the slower mzStar program.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('pymzML', 'MS:1001914',
                 ('Python module to interface mzML Data.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('conversion software', 'MS:1002333',
                 ('Computer software primarily designed to convert data '
                  'represented in one format to another format, sometimes with '
                  'minor data alterations in the process.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('preprocessing software', 'MS:1002386',
                 ('Preprocessing software.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('postprocessing software', 'MS:1002414',
                 ('Postprocessing software.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('PepFinder', 'MS:1002524',
                 ('Thermo Scientific PepFinder BioPharma analysis software.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('Lipid Data Analyzer', 'MS:1002965',
                 ('Lipid Data Analyzer software for lipid quantification.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LipidHunter', 'MS:1002967',
                 ('Software for identification of phospholipids by high- '
                  'throughput processing of LC-MS and shotgun lipidomics '
                  'datasets.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LipidXplorer', 'MS:1002968',
                 ('Software for consensual cross-platform lipidomics.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LipidMatch', 'MS:1002969',
                 ('An automated workflow for rule-based lipid identification '
                  'using untargeted high-resolution tandem mass spectrometry '
                  'data.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('Greazy', 'MS:1002970',
                 ('Open-source software for automated phospholipid tandem mass '
                  'spectrometry identification.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LipidBlast', 'MS:1002971',
                 ('LC-MS-based lipidomics and automated identification of '
                  'lipids using the LipidBlast in-silico MS/MS library.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('Lipid-Pro', 'MS:1002972',
                 ('A computational lipid identification solution for untargeted '
                  'lipidomics on data-independent acquisition tandem mass '
                  'spectrometry platforms.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LipidFinder', 'MS:1002973',
                 ('A computational workflow for the discovery of lipids for the '
                  'identification of eicosanoid-phosphoinositides in platelets.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LipiDex', 'MS:1002974',
                 ('An integrated software package for high-confidence lipid '
                  'identification.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LIQUID', 'MS:1002975',
                 ('An-open source software for identifying lipids in LC-MS/MS- '
                  'based lipidomics data.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('ALEX', 'MS:1002976',
                 ('Analysis of lipid experiments, a calculator for m/z values '
                  'of intact lipid molecules (MS1).'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('ALEX123', 'MS:1002977',
                 ('Analysis of lipid experiments 123, a calculator with m/z '
                  'values of intact lipid molecules (MS1) and their fragment '
                  'ions at the MS2 and MS3 level.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LIMSA', 'MS:1002978',
                 ('Software tool for the quantitative analysis of mass '
                  'spectrometric lipidome data.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LOBSTAHS', 'MS:1002979',
                 ('Adduct-Based lipidomics software for the discovery and '
                  'identification of oxidative stress biomarkers.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('LipidQA', 'MS:1002980',
                 ('Lipid qualitative/quantitative analysis software for '
                  'identification and quantitation of complex lipid molecular '
                  'species.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('ms_deisotope', 'MS:1002990',
                 ('ms_deisotope, a library for deisotoping and charge state '
                  'deconvolution of mass spectra.'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('LPPtiger', 'MS:1003010',
                 ('Software for lipidome-specific prediction and identification '
                  'of oxidized phospholipids from LC-MS datasets.'),
                 'software',
                 ['lipidomics analysis software', 'data processing software', 'small molecule analysis software', 'software', 'analysis software']),
    SoftwareName('MS-DIAL', 'MS:1003082',
                 ('Data processing software for untargeted metabolomics and '
                  'lipidomics that supports multiple instruments and MS '
                  'vendors.'),
                 'software',
                 ['small molecule analysis software', 'data processing software', 'analysis software', 'software']),
    SoftwareName('ThermoRawFileParser', 'MS:1003145',
                 ('Cross-platform software to convert Thermo RAW files to a '
                  'number of open formats." [DOI:10.1021/acs.jproteome.9b00328'),
                 'software',
                 ['data processing software', 'software']),
    SoftwareName('MALDI Solutions Microbial Identification', 'MS:1001878',
                 ('Shimadzu Biotech software for data acquisition, processing, '
                  'and analysis.'),
                 'software',
                 ['MALDI Solutions', 'acquisition software', 'analysis software', 'data processing software', 'Shimadzu Corporation software', 'software']),
    SoftwareName('ProteoWizard msconvert', 'MS:1002205',
                 ('Converts, filters, and processes mass spectrometry data in '
                  'variety of formats.'),
                 'software',
                 ['ProteoWizard software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('ProteoWizard idconvert', 'MS:1002206',
                 ('Converts, filters, and processes identifications from '
                  'shotgun proteomics experiments.'),
                 'software',
                 ['ProteoWizard software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('ProteoWizard chainsaw', 'MS:1002207',
                 ('Filters and processes protein sequence databases.'),
                 'software',
                 ['ProteoWizard software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('ProteoWizard msaccess', 'MS:1002208',
                 ('Filters, processes, and displays mass spectrometry data in a '
                  'variety of ways.'),
                 'software',
                 ['ProteoWizard software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('ProteoWizard SeeMS', 'MS:1002209',
                 ('An interactive GUI application to view and filter mass '
                  'spectrometry data in a variety of formats.'),
                 'software',
                 ['ProteoWizard software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('BaselineFilter', 'MS:1000753',
                 ('Removes the baseline from profile spectra using a top-hat '
                  'filter.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('DBExporter', 'MS:1000754',
                 ('Exports data from an OpenMS database to a file.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('DBImporter', 'MS:1000755',
                 ('Imports data to an OpenMS database.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('FileConverter', 'MS:1000756',
                 ('Converts between different MS file formats.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('FileFilter', 'MS:1000757',
                 ('Extracts or manipulates portions of data from peak, feature '
                  'or consensus feature files.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('FileMerger', 'MS:1000758',
                 ('Merges several MS files into one file.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('InternalCalibration', 'MS:1000759',
                 ('Applies an internal calibration.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MapAligner', 'MS:1000760',
                 ('OBSOLETE Corrects retention time distortions between maps.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('MapNormalizer', 'MS:1000761',
                 ('Normalizes peak intensities in an MS run.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('NoiseFilter', 'MS:1000762',
                 ('OBSOLETE Removes noise from profile spectra by using '
                  'different smoothing techniques.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('PeakPicker', 'MS:1000763',
                 ('OBSOLETE Finds mass spectrometric peaks in profile mass '
                  'spectra.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('Resampler', 'MS:1000764',
                 ('Transforms an LC/MS map into a resampled map or a png image.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('SpectraFilter', 'MS:1000765',
                 ('OBSOLETE Applies a filter to peak spectra.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOFCalibration', 'MS:1000766',
                 ('Applies time of flight calibration.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP noise filter', 'MS:1002131',
                 ('Noise filter component of the TOPP software.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP peak picker', 'MS:1002134',
                 ('Peak picker component of the TOPP software.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP spectra filter', 'MS:1002137',
                 ('Spectra filter component of the TOPP software.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP map aligner', 'MS:1002147',
                 ('Map aligner component of the TOPP software.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP DTAExtractor', 'MS:1002154',
                 ('Extracts spectra of an MS run file to several files in DTA '
                  'format.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP IDMerger', 'MS:1002155',
                 ('Merges several protein/peptide identification files into one '
                  'file.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP IDFileConverter', 'MS:1002156',
                 ('Converts identification engine file formats.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraMerger', 'MS:1002157',
                 ('Merges spectra from an LC/MS map, either by precursor or by '
                  'RT blocks.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP MzTabExporter', 'MS:1002158',
                 ('Exports various XML formats to an mzTab file.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP MassTraceExtractor', 'MS:1002159',
                 ('Annotates mass traces in centroided LC/MS maps.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP PrecursorMassCorrector', 'MS:1002160',
                 ('Correct the precursor entries of tandem MS scans.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP HighResPrecursorMassCorrector', 'MS:1002161',
                 ('Performs precursor mz correction on centroided high '
                  'resolution data.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP AdditiveSeries', 'MS:1002162',
                 ('Computes an additive series to quantify a peptide in a set '
                  'of samples.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP Decharger', 'MS:1002163',
                 ('Decharges and merges different feature charge variants of '
                  'the same chemical entity.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP EICExtractor', 'MS:1002164',
                 ('Quantifies signals at given positions in (raw or picked) '
                  'LC/MS maps.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP feature finder', 'MS:1002165',
                 ('Feature finder component of the TOPP software.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP ProteinQuantifier', 'MS:1002171',
                 ('Computes protein abundances from annotated feature/consensus '
                  'maps.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP ConsensusMapNormalizer', 'MS:1002172',
                 ('Normalizes maps of one consensus XML file (after linking).'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP MapRTTransformer', 'MS:1002173',
                 ('Applies retention time transformations to maps.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP feature linker', 'MS:1002174',
                 ('Feature linker component of the TOPP software.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP CompNovo', 'MS:1002178',
                 ('Performs a peptide/protein identification with the CompNovo '
                  'engine.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP CompNovoCID', 'MS:1002179',
                 ('Performs a peptide/protein identification with the CompNovo '
                  'engine in collision-induced dissociation (CID) mode.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP software adaptor', 'MS:1002180',
                 ('Software adaptor to an external program in the TOPP '
                  'software.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpecLibSearcher', 'MS:1002187',
                 ('Identifies peptide MS2 spectra by spectral matching with a '
                  'searchable spectral library.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP ConsensusID', 'MS:1002188',
                 ('Computes a consensus identification from peptide '
                  'identifications of several identification engines.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP IDConflictResolver', 'MS:1002189',
                 ('Resolves ambiguous annotations of features with peptide '
                  'identifications.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP IDFilter', 'MS:1002190',
                 ('Filters results from protein or peptide identification '
                  'engines based on different criteria.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP IDMapper', 'MS:1002191',
                 ('Assigns protein/peptide identifications to feature or '
                  'consensus features.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP IDPosteriorErrorProbability', 'MS:1002192',
                 ('Estimates posterior error probabilities using a mixture '
                  'model.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP IDRTCalibration', 'MS:1002193',
                 ('Calibrate Retention times of peptide hits to standards.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP PeptideIndexer', 'MS:1002194',
                 ('Refreshes the protein references for all peptide hits.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP PrecursorIonSelector', 'MS:1002195',
                 ('A tool for precursor ion selection based on identification '
                  'results.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP MRMMapper', 'MS:1002196',
                 ('MRMMapper maps measured chromatograms (mzML) and the '
                  'transitions used (TraML).'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP OpenSwath component', 'MS:1002197',
                 ('OpenSwath component of the TOPP software.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP ProteinInference', 'MS:1002203',
                 ('Infer proteins from a list of (high-confidence) peptides.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP FalseDiscoveryRate', 'MS:1002204',
                 ('Estimates the false discovery rate on peptide and protein '
                  'level using decoy searches.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('OpenXQuest', 'MS:1002673',
                 ('Cross-Linking MS search engine.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('OpenPepXL', 'MS:1003023',
                 ('Cross-Linking MS search engine.'),
                 'software',
                 ['TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('mzidLib:Omssa2Mzid', 'MS:1002238',
                 ('A converter for OMSSA OMX to mzIdentML.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('mzidLib:Tandem2Mzid', 'MS:1002239',
                 ('A converter for Tandem XML to mzIdentML.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('mzidLib:Csv2Mzid', 'MS:1002240',
                 ('A converter for CSV files (following OMSSA CSV style) to '
                  'mzIdentML.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('mzidLib:ProteoGrouper', 'MS:1002241',
                 ('A generic and parameterizable protein inference algorithm '
                  'for mzIdentML files.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('mzidLib:Thresholder', 'MS:1002242',
                 ('A routine for keeping only identifications passing a given '
                  'threshold or setting passThreshold to true or false for '
                  'SpectrumIdentificationItem or ProteinDetectionHypothesis in '
                  'mzIdentML files.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('mzidLib:Perform emPAI on mzid', 'MS:1002243',
                 ('A routine for adding emPAI quantitative values to an '
                  'mzIdentML file.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('mzidLib:FalseDiscoveryRate', 'MS:1002244',
                 ('A routine for calculating local FDR, q-value and FDRScore '
                  'for mzIdentML files, based on a decoy search.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('mzidLib:Mzidentml2Csv', 'MS:1002245',
                 ('A tool for converting mzIdentML files to CSV format.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('mzidLib:CombineSearchEngines', 'MS:1002246',
                 ('A tool for combining results analysed in parallel in two or '
                  'three search engines into a single mzIdentML file.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('mzidLib:InsertMetaDataFromFasta', 'MS:1002247',
                 ('A tool for adding additional meta data from a FASTA file to '
                  'DBSequence entries (sequence and description) in mzIdentML '
                  'files.'),
                 'software',
                 ['mzidLib', 'analysis software', 'software']),
    SoftwareName('PeptideProphet', 'MS:1002287',
                 ('A program in the TPP that calculates PSM probabilities for '
                  'MS2 proteomics data searched with any of the supported '
                  'sequence or spectral library search engines via the pepXML '
                  'format." [PMID:12403597'),
                 'software',
                 ['Trans-Proteomic Pipeline software', 'analysis software', 'software']),
    SoftwareName('iProphet', 'MS:1002288',
                 ('A program in the TPP that calculates distinct peptide '
                  'probabilities based on several lines of corroborating '
                  'evidence including search results from multiple search '
                  'engines via the pepXML format.'),
                 'software',
                 ['Trans-Proteomic Pipeline software', 'analysis software', 'software']),
    SoftwareName('ProteinProphet', 'MS:1002289',
                 ('A program in the TPP that calculates protein-level '
                  'probabilities based on input PSM or peptide-level '
                  'probabilities from PeptideProphet or iProphet. The output is '
                  'written in the protXML format.'),
                 'software',
                 ['Trans-Proteomic Pipeline software', 'analysis software', 'software']),
    SoftwareName('XPRESS', 'MS:1002290',
                 ('A program in the TPP that calculates PSM-level abundances '
                  'based on 2-channel isotope-labelled data such as ICAT, '
                  'SILAC, etc.'),
                 'software',
                 ['Trans-Proteomic Pipeline software', 'analysis software', 'software']),
    SoftwareName('Libra', 'MS:1002291',
                 ('A program in the TPP that calculates PSM, peptide, and '
                  'protein-level abundances based on N-channel isobaric label '
                  'peptide data such as iTRAQ, TMT, etc.'),
                 'software',
                 ['Trans-Proteomic Pipeline software', 'analysis software', 'software']),
    SoftwareName('PTMProphet', 'MS:1002292',
                 ('A program in the TPP that calculates PTM localization '
                  'probabilities by re-analyzing the peaks that are available '
                  'to distinguish between possible modification sites.'),
                 'software',
                 ['Trans-Proteomic Pipeline software', 'analysis software', 'software']),
    SoftwareName('ASAPRatio', 'MS:1002574',
                 ('A program in the TPP that calculates PSM, peptide, and '
                  'protein-level abundances based on 2-channel isotope-labelled '
                  'data such as ICAT, SILAC, etc.'),
                 'software',
                 ['Trans-Proteomic Pipeline software', 'analysis software', 'software']),
    SoftwareName('Progenesis QI', 'MS:1002879',
                 ('Metabolomics analysis software for LC-MS data from Nonlinear '
                  'Dynamics.'),
                 'software',
                 ['small molecule analysis software', 'analysis software', 'software']),
    SoftwareName('Compound Discoverer', 'MS:1002880',
                 ('Metabolomics analysis software from Thermo Fisher '
                  'Scientific.'),
                 'software',
                 ['small molecule analysis software', 'analysis software', 'software']),
    SoftwareName('MyCompoundID', 'MS:1002881',
                 ('Metabolite identification tool MyCompoundID." [PSI:PI'),
                 'software',
                 ['small molecule analysis software', 'analysis software', 'software']),
    SoftwareName('lipidomics analysis software', 'MS:1002964',
                 ('Lipidomics analysis software.'),
                 'software',
                 ['small molecule analysis software', 'analysis software', 'software']),
    SoftwareName('ProCon', 'MS:1002334',
                 ('Java software designed to convert one of several proteomics '
                  'identification results formats into mzIdentML or PRIDE XML." '
                  '[PSI:PI'),
                 'software',
                 ['conversion software', 'data processing software', 'software']),
    SoftwareName('PRIDE Converter2', 'MS:1002335',
                 ('Java software designed to convert one of several proteomics '
                  'identification results formats into PRIDE XML.'),
                 'software',
                 ['conversion software', 'data processing software', 'software']),
    SoftwareName('python-psims', 'MS:1002991',
                 ('python-psims, a library for generating mzML and mzIdentML.'),
                 'software',
                 ['conversion software', 'data processing software', 'software']),
    SoftwareName('TOPP NoiseFilterGaussian', 'MS:1002132',
                 ('Removes noise from profile spectra by using a gaussian '
                  'smoothing.'),
                 'software',
                 ['TOPP noise filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP NoiseFilterSGolay', 'MS:1002133',
                 ('Removes noise from profile spectra by using a Savitzky-Golay '
                  'smoothing.'),
                 'software',
                 ['TOPP noise filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP PeakPickerHiRes', 'MS:1002135',
                 ('Finds mass spectrometric peaks in high-resoluted profile '
                  'mass spectra.'),
                 'software',
                 ['TOPP peak picker', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP PeakPickerWavelet', 'MS:1002136',
                 ('Finds mass spectrometric peaks with a wavelet algorithm in '
                  'low-resoluted profile mass spectra.'),
                 'software',
                 ['TOPP peak picker', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraFilterBernNorm', 'MS:1002138',
                 ('Applies a Bern et al normalization to peak spectra." '
                  '[PMID:15262780'),
                 'software',
                 ['TOPP spectra filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraFilterMarkerMower', 'MS:1002139',
                 ('Applies a filter to peak spectra for marked peaks.'),
                 'software',
                 ['TOPP spectra filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraFilterNLargest', 'MS:1002140',
                 ('Retains the n largest peaks of a peak spectra.'),
                 'software',
                 ['TOPP spectra filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraFilterNormalizer', 'MS:1002141',
                 ('Applies a TIC/maximal intensity normalization to peak '
                  'spectra.'),
                 'software',
                 ['TOPP spectra filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraFilterParentPeakMower', 'MS:1002142',
                 ('Filters putative unfragmented precursor ions from tandem '
                  'spectra.'),
                 'software',
                 ['TOPP spectra filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraFilterScaler', 'MS:1002143',
                 ('Applies a filter to peak spectra after intensity scaling '
                  'according to rank.'),
                 'software',
                 ['TOPP spectra filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraFilterSqrtMower', 'MS:1002144',
                 ('Applies a filter to peak spectra after intensity scaling to '
                  'the square root.'),
                 'software',
                 ['TOPP spectra filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraFilterThresholdMower', 'MS:1002145',
                 ('Applies a filter of peaks below a given threshold to peak '
                  'spectra.'),
                 'software',
                 ['TOPP spectra filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP SpectraFilterWindowMower', 'MS:1002146',
                 ('Applies a filter of the largest peaks in a sliding window '
                  'over a peak spectrum.'),
                 'software',
                 ['TOPP spectra filter', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP MapAlignerIdentification', 'MS:1002148',
                 ('Corrects retention time distortions between maps based on '
                  'common peptide identifications.'),
                 'software',
                 ['TOPP map aligner', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP MapAlignerPoseClustering', 'MS:1002149',
                 ('Corrects retention time distortions between maps using a '
                  'pose clustering approach.'),
                 'software',
                 ['TOPP map aligner', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP MapAlignerSpectrum', 'MS:1002150',
                 ('Corrects retention time distortions between maps by spectrum '
                  'alignment.'),
                 'software',
                 ['TOPP map aligner', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP FeatureFinderCentroided', 'MS:1002166',
                 ('Detects two-dimensional features in centroided LC-MS data.'),
                 'software',
                 ['TOPP feature finder', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP FeatureFinderRaw', 'MS:1002167',
                 ('Detects two-dimensional features in uncentroided LC-MS data.'),
                 'software',
                 ['TOPP feature finder', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP FeatureFinderIsotopeWavelet', 'MS:1002168',
                 ('Detects two-dimensional features in uncentroided LC-MS data '
                  'with a wavelet algorithm.'),
                 'software',
                 ['TOPP feature finder', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP FeatureFinderMetabo', 'MS:1002169',
                 ('Detects two-dimensional features in centroided LC-MS data of '
                  'metabolites.'),
                 'software',
                 ['TOPP feature finder', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP FeatureFinderMRM', 'MS:1002170',
                 ('Quantifies features LC-MS/MS MRM data.'),
                 'software',
                 ['TOPP feature finder', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP FeatureLinkerLabeled', 'MS:1002175',
                 ('Groups corresponding isotope-labeled features in a feature '
                  'map.'),
                 'software',
                 ['TOPP feature linker', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP FeatureLinkerUnlabeled', 'MS:1002176',
                 ('Groups corresponding features from multiple maps.'),
                 'software',
                 ['TOPP feature linker', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP FeatureLinkerUnlabeledQT', 'MS:1002177',
                 ('Groups corresponding features from multiple maps using a '
                  'quality threshold clustering approach.'),
                 'software',
                 ['TOPP feature linker', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP InspectAdapter', 'MS:1002181',
                 ('Identifies MS2 spectra using the external program Inspect.'),
                 'software',
                 ['TOPP software adaptor', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP MascotAdapter', 'MS:1002182',
                 ('Identifies MS2 spectra using the external program Mascot.'),
                 'software',
                 ['TOPP software adaptor', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP MascotAdapterOnline', 'MS:1002183',
                 ('Identifies MS2 spectra using the online version of the '
                  'external program Mascot.'),
                 'software',
                 ['TOPP software adaptor', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP OMSSAAdapter', 'MS:1002184',
                 ('Identifies MS2 spectra using the external program OMSSA.'),
                 'software',
                 ['TOPP software adaptor', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP PepNovoAdapter', 'MS:1002185',
                 ('Identifies MS2 spectra using the external program PepNovo.'),
                 'software',
                 ['TOPP software adaptor', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP XTandemAdapter', 'MS:1002186',
                 ('Identifies MS2 spectra using the external program XTandem.'),
                 'software',
                 ['TOPP software adaptor', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP OpenSwathAnalyzer', 'MS:1002198',
                 ('Picks peaks and finds features in an SRM experiment.'),
                 'software',
                 ['TOPP OpenSwath component', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP OpenSwathChromatogramExtractor', 'MS:1002199',
                 ('Extract chromatograms (XIC) from a MS2 map file.'),
                 'software',
                 ['TOPP OpenSwath component', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP OpenSwathDecoyGenerator', 'MS:1002200',
                 ('Generates decoys according to different models for a '
                  'specific TraML.'),
                 'software',
                 ['TOPP OpenSwath component', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP OpenSwathFeatureXMLToTSV', 'MS:1002201',
                 ('Converts a featureXML to a mProphet tsv (tab separated '
                  'values).'),
                 'software',
                 ['TOPP OpenSwath component', 'TOPP software', 'analysis software', 'data processing software', 'software']),
    SoftwareName('TOPP OpenSwathRTNormalizer', 'MS:1002202',
                 ('Generates a transformation file for retention time space '
                  'into normalized space.'),
                 'software',
                 ['TOPP OpenSwath component', 'TOPP software', 'analysis software', 'data processing software', 'software']),
])
# [[[end]]]


software_names_by_name = {c.name: c for c in software_names}


def software_name(name):
    try:
        return software_names_by_name[name]
    except KeyError:
        return SoftwareName(name, name, name, name, [name])


__all__ = [
    "SoftwareName", "Software", "software_names",
    "software_name"
]




from functools import total_ordering
from collections import namedtuple, Counter


def isclose(a, b):
    return abs(a - b) < 1e-3


CHARGE_UNKNOWN = 0
OUT_OF_RANGE_INT = 999


def intensity_ratio_function(peak1, peak2):
    ratio = peak1.intensity / float(peak2.intensity)
    if ratio >= 5:
        return -4
    elif 2.5 <= ratio < 5:
        return -3
    elif 1.7 <= ratio < 2.5:
        return -2
    elif 1.3 <= ratio < 1.7:
        return -1
    elif 1.0 <= ratio < 1.3:
        return 0
    elif 0.8 <= ratio < 1.0:
        return 1
    elif 0.6 <= ratio < 0.8:
        return 2
    elif 0.4 <= ratio < 0.6:
        return 3
    elif 0.2 <= ratio < 0.4:
        return 4
    elif 0. <= ratio < 0.2:
        return 5


class DeltaFeature(object):
    def __init__(self, offset, tolerance, intensity_ratio, charge_from, charge_to, name=None):
        self.offset = offset
        self.tolerance = tolerance
        self.intensity_ratio = intensity_ratio
        self.charge_from = charge_from
        self.charge_to = charge_to
        self._hash = hash((self.offset, self.intensity_ratio, self.from_charge,
                           self.to_charge))

    def test(self, peak1, peak2):
        if (self.intensity_ratio == OUT_OF_RANGE_INT or
                intensity_ratio_function(peak1, peak2) == self.intensity_ratio) and\
            ((self.from_charge == OUT_OF_RANGE_INT and self.to_charge == OUT_OF_RANGE_INT) or
             (self.from_charge == peak1.charge and self.to_charge == peak2.charge)):

            return abs((peak1.neutral_mass + self.offset - peak2.neutral_mass) / peak2.neutral_mass) <= self.tolerance
        return False

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        v = abs(self.offset - other.offset) <= 1e-3
        if not v:
            return v
        v = self.intensity_ratio == other.intensity_ratio
        if not v:
            return v
        v = self.from_charge == other.from_charge
        if not v:
            return v
        v = self.to_charge == other.to_charge
        if not v:
            return v
        return True

    def __lt__(self, other):
        eq_count = 0
        v = self.intensity_ratio <= other.intensity_ratio
        eq_count += self.intensity_ratio == other.intensity_ratio
        if not v:
            return v
        v = self.from_charge <= other.from_charge
        eq_count += self.from_charge == other.from_charge
        if not v:
            return v
        v = self.to_charge <= other.to_charge
        eq_count += self.to_charge == other.to_charge
        if not v:
            return v
        if eq_count == 3:
            return False
        return True

    def __gt__(self, other):
        eq_count = 0
        v = self.intensity_ratio >= other.intensity_ratio
        eq_count += self.intensity_ratio == other.intensity_ratio
        if not v:
            return v
        v = self.from_charge >= other.from_charge
        eq_count += self.from_charge == other.from_charge
        if not v:
            return v
        v = self.to_charge >= other.to_charge
        eq_count += self.to_charge == other.to_charge
        if not v:
            return v
        if eq_count == 3:
            return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def __call__(self, peak1, peak2, structure=None):
        return self.test(peak1, peak2)

    def find_matches(self, peak, peak_list):
        matches = []
        for peak2 in peak_list.all_peaks_for(peak.neutral_mass + self.offset, self.tolerance):
            if self(peak, peak2) and peak is not peak2:
                matches.append(peak2)
        return matches

    def is_valid_match(self, peak, match, labeled_peaks):
        return match in labeled_peaks

    def specialize(self, from_charge, to_charge, intensity_ratio):
        return self.__class__(
            self.offset, self.tolerance, self.name, intensity_ratio,
            from_charge, to_charge)

    def unspecialize(self):
        return self.__class__(
            self.offset, self.tolerance, self.name, OUT_OF_RANGE_INT,
            OUT_OF_RANGE_INT, OUT_OF_RANGE_INT)

    def _get_display_fields(self):
        fields = {}
        fields['offset'] = self.offset
        if self.from_charge != OUT_OF_RANGE_INT:
            fields["from_charge"] = self.from_charge
        if self.to_charge != OUT_OF_RANGE_INT:
            fields['to_charge'] = self.to_charge
        if self.intensity_ratio != OUT_OF_RANGE_INT:
            fields["intensity_ratio"] = self.intensity_ratio
        terms = []
        for k, v in fields.items():
            if isinstance(v, int):
                terms.append("%s=%d" % (k, v))
            elif isinstance(v, float):
                terms.append("%s=%0.4f" % (k, v))
            else:
                terms.append("%s=%r" % (k, v))
        return terms

    def __repr__(self):
        terms = self._get_display_fields()
        return "{}(name={!r}, {})".format(
            self.__class__.__name__, self.name, ", ".join(terms))


def feature_function_estimator(labeled_scans, feature_function, tolerance=2e-5, track_relations=True, verbose=False):
    total_on_series_satisfied = 0.
    total_off_series_satisfied = 0.
    total_on_series = 0.
    total_off_series = 0.
    peak_relations = []
    for i_scan, labeled_scan in enumerate(labeled_scans):
        if verbose and i_scan % 1000 == 0:
            print(i_scan)
        peaks = labeled_scan.deconvoluted_peak_set
        related = []
        labeled_peaks = labeled_scan.annotations['labeled_peaks']
        for peak in peaks:
            is_on_series = peak in labeled_peaks
            matches = feature_function.find_matches(peak, peaks)
            for match in matches:
                if peak is match:
                    continue
                if track_relations:
                    pr = PeakRelation(
                        peak, match, feature_function, intensity_ratio_function(peak, match))
                    related.append(pr)
                is_match_expected = feature_function.is_valid_match(peak, match, labeled_peaks)
                if is_on_series and is_match_expected:
                    total_on_series_satisfied += 1
                else:
                    total_off_series_satisfied += 1
            if is_on_series:
                total_on_series += 1
            else:
                total_off_series += 1
        if len(related) > 0 and track_relations:
            peak_relations.append((labeled_scan, related))

    total_on_series_satisfied_normalized = total_on_series_satisfied / \
        max(total_on_series, 1)
    total_off_series_satisfied_normalized = total_off_series_satisfied / \
        max(total_off_series, 1)

    return FittedFeature(feature_function, total_on_series_satisfied_normalized,
                         total_off_series_satisfied_normalized, peak_relations,
                         on_count=total_on_series_satisfied,
                         off_count=total_off_series_satisfied)


class FittedFeatureBase(object):
    def find_matches(self, peak, peak_list):
        return self.feature.find_matches(peak, peak_list)

    def is_valid_match(self, from_peak, to_peak, labeled_peaks):
        return self.feature.is_valid_match(from_peak, to_peak, labeled_peaks)

    def _feature_probability(self, p=0.5):
        return (p * self.on_series) / (
            (p * self.on_series) + ((1 - p) * self.off_series))


@total_ordering
class FittedFeature(FittedFeatureBase):
    def __init__(self, feature, on_series, off_series, relations=None, on_count=0, off_count=0):
        if relations is None:
            relations = []
        self.feature = feature

        # forward these attributes directly rather than using a property
        # to avoid overhead
        self.from_charge = feature.from_charge
        self.to_charge = feature.to_charge

        # mu
        self.on_series = on_series
        # v
        self.off_series = off_series

        # tracking purposes only
        self.on_count = on_count
        self.off_count = off_count
        self.relations = relations

    @property
    def _total_on_series(self):
        if self.on_series == 0:
            return 1
        return self.on_count / self.on_series

    @property
    def _total_off_series(self):
        if self.off_series == 0:
            return 1
        return self.off_count / self.off_series

    @property
    def name(self):
        return self.feature.name

    @property
    def observations(self):
        count = (self.on_count + self.off_count)
        if count > 0:
            return count
        count = len(list(self.peak_relations()))
        if count > 0:
            return count
        return 0

    def __hash__(self):
        return hash((self.feature, self.expected))

    def __eq__(self, other):
        v = self.feature == other.feature and self.expected == other.expected
        if not v:
            return v
        v = isclose(self.on_series, other.on_series) and isclose(
            self.off_series, other.off_series)
        if not v:
            return v
        return True

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        v = self.feature < other.feature
        if not v:
            return False
        v = self.on_series < other.on_series
        if not v:
            return False
        v = self.off_series < other.off_series
        if not v:
            return False
        return True

    def __repr__(self):
        temp = ("<FittedFeature {feature.name}, {terms} u:{on_series:0.4g}"
                " v:{off_series:0.4g} @ {series} {count_relations}>")
        return temp.format(
            feature=self.feature,
            terms=', '.join(map(str, self.feature._get_display_fields())),
            on_series=self.on_series, off_series=self.off_series,
            series=self.expected, count_relations=self.on_count + self.off_count)

    def charge_relations(self):
        counter = Counter()
        for rel in self.peak_relations(False):
            counter[rel.from_charge, rel.to_charge] += 1
        return counter

    def intensity_ratio(self):
        counter = Counter()
        for rel in self.peak_relations(False):
            counter[intensity_ratio_function(rel.from_peak, rel.to_peak)] += 1
        return counter

    def charge_intensity_ratio(self):
        counter = Counter()
        for rel in self.peak_relations(False):
            counter[(rel.from_charge, rel.to_charge),
                    intensity_ratio_function(rel.from_peak, rel.to_peak)] += 1
        return counter

    def peak_relations(self, include_noise=True):
        for _, peak_relations in self.relations:
            for pr in peak_relations:
                if not include_noise and not pr.expected:
                    continue
                yield pr

    def partitions(self, minimum_count=10):
        counter = Counter()
        on_counter = Counter()
        off_counter = Counter()
        for rel in self.peak_relations(True):
            fs = FeatureSpecialization(
                from_charge=rel.from_charge,
                to_charge=rel.to_charge,
                intensity_ratio=rel.intensity_ratio,
            )
            counter[fs] += 1
            if rel.expected == self.expected:
                on_counter[fs] += 1
            else:
                off_counter[fs] += 1

        counter = {k: (v, on_counter[k], off_counter[k])
                   for k, v in counter.items() if v >= minimum_count}
        return counter

    def specialize(self, minimum_count=10):
        counter = self.partitions(minimum_count)
        total_off_series = self._total_off_series
        total_on_series = self._total_on_series
        specialized_features = []
        for params, counts in counter.items():
            feature = self.feature.specialize(
                intensity_ratio=params.intensity_ratio,
                from_charge=params.from_charge,
                to_charge=params.to_charge)
            total, on, off = counts
            fit = FittedFeature(
                feature, on / total_on_series, off / total_off_series,
                None, on, off)
            specialized_features.append(fit)
        return specialized_features

    def __call__(self, peak1, peak2, structure=None):
        return self.feature(peak1, peak2, structure)

    def pack(self):
        self.relations = []
        return self

    def to_json(self):
        d = {}
        d['on_series'] = self.on_series
        d['on_count'] = self.on_count
        d['off_series'] = self.off_series
        d['off_count'] = self.off_count
        d['feature'] = self.feature.to_json()
        return d

    @classmethod
    def from_json(cls, d):
        feature = DeltaFeature.from_json(d['feature'])
        inst = cls(
            feature, d["on_series"],
            d["off_series"], relations=None, on_count=d['on_count'],
            off_count=d['off_count']
        )
        return inst


FeatureSpecialization = namedtuple("FeatureSpecialization", [
                                   "from_charge", "to_charge", "intensity_ratio"])


class PeakRelation(object):
    __slots__ = ["from_peak", "to_peak", "feature",
                 "intensity_ratio", "expected", "from_charge",
                 "to_charge"]

    def __init__(self, from_peak, to_peak, feature, intensity_ratio=None, expected=None):
        if intensity_ratio is None:
            intensity_ratio = intensity_ratio_function(from_peak, to_peak)
        self.from_peak = from_peak
        self.to_peak = to_peak
        self.feature = feature
        self.intensity_ratio = intensity_ratio
        self.from_charge = from_peak.charge
        self.to_charge = to_peak.charge
        self.expected = expected

    def __reduce__(self):
        return self.__class__, (self.from_peak, self.to_peak, self.feature, self.intensity_ratio, self.expected)

    def __repr__(self):
        template = "<PeakRelation {s.from_peak.neutral_mass}({s.from_charge}) ->" +\
            " {s.to_peak.neutral_mass}({s.to_charge}) by {s.feature.name} on {s.expected}>"
        return template.format(s=self)

    def peak_key(self):
        if self.from_peak.index.neutral_mass < self.to_peak.index.neutral_mass:
            return self.from_peak, self.to_peak
        else:
            return self.to_peak, self.from_peak



"""Code for writing Mascot Generic Format files and for reading deconvoluted
generic format files.

"""

from ms_deisotope.averagine import neutral_mass, mass_charge_ratio
from ms_deisotope.peak_set import DeconvolutedPeak, DeconvolutedPeakSet
from ms_deisotope.data_source.mgf import MGFLoader, mgf as pymgf, _MGFParser

from .text import HeaderedDelimitedWriter


def _format_parameter(key, value):
    return "{0}={1}\n".format(str(key).upper(), str(value))


class MGFSerializer(HeaderedDelimitedWriter):
    """A MASCOT Generic Format writer which is deconvolution aware.

    Implements :class:`~.HeaderedDelimitedWriter, as well as allowing
    global parameters to be included at the beginning of the file.

    Attributes
    ----------
    sample_name: str
        The name of the sample. This value is currently ignored.
    started: bool
        Whether scans have been written out yet or not. Once :attr:`started` is
        true, it is no longer possible to add global parameters using :meth:`add_global_parameter`.

    """

    file_extensions = {
        "mgf",
        "mgf.gz"
    }

    def __init__(self, stream, sample_name=None, deconvoluted=True):
        super(MGFSerializer, self).__init__(stream, deconvoluted)
        self.sample_name = sample_name
        self.started = False

    def add_global_parameter(self, name, value):
        """Add a global parameter at the beginning of the file, before scans
        are written.

        Global parameters follow the same ``KEY=value`` format as scan-level
        parameters. New global parameters may not be added once a scan has been
        written.

        Parameters
        ----------
        name : str
            The parameter name. Will be made upper-case.
        value : object
            The parameter's value. Will be converted to a string.

        Raises
        ------
        ValueError:
            If scans have already been written, this method raises a :class:`ValueError`
        """
        if self.started:
            raise ValueError("Cannot add global parameter if scan data has begun being written")
        self._add_parameter(name, value)

    def _add_parameter(self, name, value):
        self.stream.write(_format_parameter(name, value).encode('utf-8'))

    def add_parameter(self, name, value):
        """Add a parameter to the current block.

        Parameters are written ``KEY=value`` format.

        Parameters
        ----------
        name : str
            The parameter name. Will be made upper-case.
        value : object
            The parameter's value. Will be converted to a string.
        """
        self._add_parameter(name, value)

    def save_scan_bunch(self, bunch, **kwargs):
        for scan in bunch.products:
            self.save_scan(scan, **kwargs)

    def format_peak_vectors(self, scan):
        """As in :class:`~.HeaderedDelimitedTextWriter` but always writes m/z, even when charge
        is known.
        """
        if self.deconvoluted:
            (neutral_mass_array, intensity_array, charge_array) = super(
                MGFSerializer, self).format_peak_vectors(scan)
            mz_array = [mass_charge_ratio(
                neutral_mass_array[i], charge_array[i]) for i in range(len(charge_array))]
        else:
            (mz_array, intensity_array, charge_array) = super(
                MGFSerializer, self).format_peak_vectors(scan)
        return (mz_array, intensity_array, charge_array)

    def write_header(self, header_dict):
        pepmass = header_dict['precursor_mz']
        charge = header_dict['precursor_charge']
        intensity = header_dict['precursor_intensity']
        self.add_parameter("pepmass", "%f %f" % (pepmass, intensity))
        polarity = header_dict['polarity']
        if polarity is None:
            polarity = 1
        try:
            self.add_parameter("charge", "%d%s" % (charge, "+" if polarity > 0 else '-'))
        except TypeError:
            pass
        self.add_parameter("title", header_dict['title'])
        self.add_parameter("rtinseconds", header_dict['scan_time'] * 60.0)

    def write_scan(self, scan_header, data_vectors):
        self.stream.write(b'BEGIN IONS\n')
        self.write_header(scan_header)
        self.write_vectors(data_vectors)
        self.stream.write(b'END IONS\n')


class ProcessedMGFLoader(MGFLoader):
    """A variant of the :class:`~.MGFLoader` that reads deconvoluted mass spectra
    with a charge value for each peak, and looks for additional annotation of the
    precursor ion.

    """

    file_extensions = {
        "mgf",
        "mgf.gz"
    }

    def __init__(self, source_file, encoding='ascii', use_index=True, ** kwargs):
        super(ProcessedMGFDeserializer, self).__init__(
            source_file, encoding, use_index, **kwargs)

    def _create_parser(self):
        if self._use_index:
            return _MGFParser(self.source_file, read_charges=True,
                              convert_arrays=1, encoding=self.encoding)
        else:
            return pymgf.MGF(self.source_file, read_charges=True,
                             convert_arrays=1, encoding=self.encoding)

    def _build_peaks(self, scan):
        mz_array = scan['m/z array']
        intensity_array = scan["intensity array"]
        charge_array = scan['charge array']
        return build_deconvoluted_peak_set_from_arrays(mz_array, intensity_array, charge_array)

    def _make_scan(self, data):
        scan = super(ProcessedMGFDeserializer, self)._make_scan(data)
        scan.peak_set = None
        scan.deconvoluted_peak_set = self._build_peaks(scan._data)
        return scan.pack(bind=True)

    def _precursor_information(self, scan):
        pinfo = super(ProcessedMGFDeserializer, self)._precursor_information(scan)
        defaulted = pinfo.defaulted
        orphan = pinfo.orphan
        pinfo.default()
        pinfo.defaulted = defaulted
        pinfo.orphan = orphan
        return pinfo


ProcessedMGFDeserializer = ProcessedMGFLoader


def build_deconvoluted_peak_set_from_arrays(mz_array, intensity_array, charge_array):
    peaks = []
    for i in range(len(mz_array)):
        peak = DeconvolutedPeak(
            neutral_mass(mz_array[i], charge_array[i]), intensity_array[i], charge_array[i],
            intensity_array[i], i, 0)
        peaks.append(peak)
    peak_set = DeconvolutedPeakSet(peaks)
    peak_set.reindex()
    return peak_set


try:
    _build_deconvoluted_peak_set_from_arrays = build_deconvoluted_peak_set_from_arrays
    from ms_deisotope._c.utils import build_deconvoluted_peak_set_from_arrays
except ImportError:
    pass



import os
import io

from typing import Dict, Union, List, Type, Callable

from .mzml import MzMLLoader
from .mzxml import MzXMLLoader
from .mgf import MGFLoader
from .scan.loader import RandomAccessScanSource, ScanIterator
from . import _compression


class FormatGuesser(object):
    guessers: List[Callable[[Union[io.IOBase, os.PathLike, str]], Type[RandomAccessScanSource]]]
    reader_types: List[Type[RandomAccessScanSource]]

    def __init__(self, guessers: List[Callable], reader_types: List[Type]):
        self.guessers = list(guessers or [])
        self.reader_types = list(reader_types or [])

    def register_type_guesser(self, reader_guesser: Callable[[Union[io.IOBase, os.PathLike, str]], Type[RandomAccessScanSource]]):
        self.guessers.append(reader_guesser)
        return reader_guesser

    def guess_type(self, file_path: Union[io.IOBase, os.PathLike, str]):
        for guesser in self.guessers:
            try:
                reader_type = guesser(file_path)
                return reader_type
            except (ValueError, IOError, ImportError, TypeError, AttributeError):
                continue
        raise ValueError("Cannot determine ScanLoader type from %r" % (file_path, ))

    def add_reader_type(self, reader: Type[RandomAccessScanSource]):
        self.reader_types.append(reader)

    def guess_from_file_extension(self, file_path: Union[io.IOBase, os.PathLike, str], ext_map: Dict[str, Type[RandomAccessScanSource]]):
        if hasattr(file_path, 'name'):
            file_path = file_path.name
        if file_path.endswith(".gz"):
            file_path = file_path[:-3]
        ext = os.path.splitext(file_path)[1].lower()
        if ext in ext_map:
            return ext_map[ext]
        else:
            raise ValueError("Cannot determine opener type from file path")

    def add_file_extension_guesser(self, ext_map):
        def guesser(file_path):
            return self.guess_from_file_extension(file_path, ext_map)

        self.register_type_guesser(guesser)
        return guesser

    def open_file(self, file_path: Union[io.IOBase, os.PathLike, str], *args, **kwargs) -> Union[RandomAccessScanSource, ScanIterator]:
        """Factory function to create an object that reads scans from
        any supported data file format. Provides both iterative and
        random access.
        """
        reader_type = self.guess_type(file_path)
        try:
            is_file = os.path.isfile(file_path)
        except TypeError:
            is_file = False
        if is_file:
            is_gz_compressed = _compression.test_gzipped(file_path)
            if is_gz_compressed:
                fobj = _compression.get_opener(file_path)
            else:
                fobj = file_path
        else:
            fobj = file_path
        return reader_type(fobj, *args, **kwargs)

    def __call__(self, file_path: Union[io.IOBase, os.PathLike], *args, **kwargs) -> Union[RandomAccessScanSource, ScanIterator]:
        return self.open_file(file_path, *args, **kwargs)


MSFileLoader = FormatGuesser([], [MzMLLoader, MzXMLLoader, MGFLoader])

register_type_guesser = MSFileLoader.register_type_guesser
guessers = MSFileLoader.guessers
reader_types = MSFileLoader.reader_types

guess_type_from_path = MSFileLoader.add_file_extension_guesser({
    ".mzml": MzMLLoader,
    ".mzxml": MzXMLLoader,
    ".mgf": MGFLoader
})


@register_type_guesser
def guess_type_from_file_sniffing(file_path):
    is_random_access_file = is_random_access(file_path)
    if is_random_access_file:
        handle = file_path
        header = handle.read(1000)
        handle.seek(0)
    else:
        with open(file_path, 'rb') as handle:
            header = handle.read(1000)

    if _compression.starts_with_gz_magic(header):
        if is_random_access_file:
            handle = _compression.GzipFile(fileobj=file_path, mode='rb')
            header = handle.read(1000)
            handle.seek(0)
        else:
            with _compression.GzipFile(file_path, mode='rb') as handle:
                header = handle.read(1000)

    if b"mzML" in header:
        return MzMLLoader
    elif b"mzXML" in header:
        return MzXMLLoader
    elif b"BEGIN IONS" in header:
        return MGFLoader
    else:
        raise ValueError("Cannot determine ScanLoader type from header")


guess_type = MSFileLoader.guess_type


def is_random_access(fp):
    return hasattr(fp, 'read') and hasattr(fp, 'seek')


try:
    from .mzmlb import MzMLbLoader, infer_reader as _infer_mzmlb, determine_if_available

    if determine_if_available():
        reader_types.append(MzMLbLoader)
        register_type_guesser(_infer_mzmlb)
except ImportError:
    pass


try:
    from .thermo_raw_net import (
        ThermoRawLoader as ThermoRawNetLoader, infer_reader as _check_is_thermo_raw_net,
        register_dll as register_thermo_net_dll)
    reader_types.append(ThermoRawNetLoader)
    register_type_guesser(_check_is_thermo_raw_net)

except ImportError:  # pragma: no cover
    def register_thermo_net_dll(*args, **kwargs):
        pass

try:
    from .thermo_raw import (
        ThermoRawLoader, infer_reader as _check_is_thermo_raw,
        register_dll as register_thermo_dll)

    reader_types.append(ThermoRawLoader)
    register_type_guesser(_check_is_thermo_raw)

except ImportError:  # pragma: no cover
    def register_thermo_dll(*args, **kwargs):
        pass


try:
    from .agilent_d import (
        AgilentDLoader,
        infer_reader as _check_is_agilent_d,
        register_dll_dir as register_agilent_dll_dir)

    reader_types.append(AgilentDLoader)
    register_type_guesser(_check_is_agilent_d)
except ImportError:  # pragma: no cover
    def register_agilent_dll_dir(*args, **kwargs):
        pass


try:
    from .masslynx import (
        MassLynxRawLoader, infer_reader as _check_mass_lynx_raw,
        register_waters_masslynx_dll)

    reader_types.append(MassLynxRawLoader)
    register_type_guesser(_check_mass_lynx_raw)
except ImportError:
    def register_waters_masslynx_dll(*args, **kwargs):
        pass



"""Thermo RAW file reading implementation using the pure .NET
RawFileReader library released in 2017.

This module provides :class:`ThermoRawLoader`, a :class:`~.RandomAccessScanSource`
implementation.

Depends upon the ``pythonnet`` project which provides the :mod:`clr`
module, enabling nearly seamless interoperation with the Common Language
Runtime.

The public interface of this module should be identical to
:mod:`ms_deisotope.data_source.thermo_raw`.

.. note::
    This interface was largely based upon the APIs that ProteoWizard used, both
    in order to understand how the Thermo libraries really worked, and to maintain
    parity with it.
"""

import sys
import os

from collections import OrderedDict
from typing import List

import numpy as np

from pyteomics.auxiliary import unitfloat

from six import string_types as basestring


from ms_peak_picker import PeakSet, PeakIndex, simple_peak

from ms_deisotope.data_source.common import (
    PrecursorInformation, ChargeNotProvided, Scan,
    ActivationInformation, MultipleActivationInformation,
    IsolationWindow, ScanDataSource, ScanEventInformation,
    ScanAcquisitionInformation, ScanWindow, RandomAccessScanSource)

from ms_deisotope.data_source._thermo_helper import (
    _InstrumentMethod, ThermoRawScanPtr, FilterString,
    _make_id, _id_template, _RawFileMetadataLoader, analyzer_map)

from ms_deisotope.data_source.metadata.activation import (
    supplemental_term_map, dissociation_methods_map)

from ms_deisotope.data_source.metadata import software
from ms_deisotope.data_source.metadata.sample import Sample
from ms_deisotope.data_source.metadata.scan_traits import FAIMS_compensation_voltage

def _try_number(string):
    try:
        x = float(string)
        return x
    except (TypeError, ValueError):
        return string


_DEFAULT_DLL_PATH = os.path.join(
    os.path.dirname(
        os.path.realpath(__file__)),
    "_vendor",
    "ThermoRawFileReader_3_0_41",
    "Libraries")


# late binding imports

Business = None
_RawFileReader = None
clr = None
NullReferenceException = Exception
Marshal = None
IntPtr = None
Int64 = None


def _open_raw_file(path):
    if not _test_dll_loaded():
        register_dll()
        if not _test_dll_loaded():
            raise ValueError(
                "Could not load .NET runtime or library. Verify `pythonnet` "
                "is installed correctly and the RawFileReader library is registered")
    raw_file = _RawFileReader.RawFileReaderAdapter.FileFactory(path)
    return raw_file


def is_thermo_raw_file(path):
    """Detect whether or not the file referenced by ``path``
    is a Thermo RAW file.

    Parameters
    ----------
    path: :class:`str`
        The path to test

    Returns
    -------
    :class:`bool`:
        Whether or not the file is a Thermo RAW file.
    """
    if not _test_dll_loaded():
        try:
            register_dll()
        except ImportError:
            return False
    with open(path, 'rb') as fh:
        lead_bytes = fh.read(32)
        decoded = lead_bytes.decode("utf-16")[1:9]
        if decoded == "Finnigan":
            try:
                source = _open_raw_file(path)
                source.SelectInstrument(Business.Device.MS, 1)
                return True
            except NullReferenceException:   # pylint: disable=broad-except
                return False
        return False


def infer_reader(path):
    """If the file referenced by ``path`` is a Thermo RAW
    file, return the callable (:class:`ThermoRawLoader`) to
    open it, otherwise raise an exception.

    Parameters
    ----------
    path: :class:`str`
        The path to test

    Returns
    -------
    :class:`type`:
        The type to use to open the file

    Raises
    ------
    :class:`ValueError`:
        If the file is not a Thermo RAW file
    """
    if is_thermo_raw_file(path):
        return ThermoRawLoader
    raise ValueError("Not Thermo Raw File")


def determine_if_available():
    """Checks whether or not the .NET-based Thermo
    RAW file reading feature is available.

    Returns
    -------
    :class:`bool`:
        Whether or not the feature is enabled.
    """
    try:
        return _register_dll([_DEFAULT_DLL_PATH])
    except (OSError, ImportError):
        return False


def _register_dll(search_paths=None):
    """Start the Common Language Runtime interop service by importing
    the :mod:`clr` module from Pythonnet, and then populate the global
    names referring to .NET entities, and finally attempt to locate the
    ThermoRawFileReader DLLs by searching alogn ``search_paths``.

    Parameters
    ----------
    search_paths: list
        The paths to check along for the ThermoRawFileReader DLL bundle.

    Returns
    -------
    :class:`bool`:
        Whether or not the .NET library successfully loaded
    """
    from ms_deisotope.config import get_config
    if search_paths is None:
        search_paths = []
    search_paths = list(search_paths)
    search_paths.append(_DEFAULT_DLL_PATH)
    # Take user-specified search paths first.
    search_paths = get_config().get('vendor_readers', {}).get('thermo-net', []) + search_paths
    global _RawFileReader, Business, clr, NullReferenceException   # pylint: disable=global-statement
    global Marshal, IntPtr, Int64   # pylint: disable=global-statement
    if _test_dll_loaded():
        return True
    try:
        import clr  # pylint: disable=redefined-outer-name
        from System import NullReferenceException  # pylint: disable=redefined-outer-name
        clr.AddReference("System.Runtime")
        clr.AddReference("System.Runtime.InteropServices")
        from System import IntPtr, Int64  # pylint: disable=redefined-outer-name
        from System.Runtime.InteropServices import Marshal  # pylint: disable=redefined-outer-name
    except ImportError:
        return False
    for path in search_paths:
        sys.path.append(path)
        try:
            clr.AddReference('ThermoFisher.CommonCore.RawFileReader')
            clr.AddReference('ThermoFisher.CommonCore.Data')
        except OSError:
            continue
        try:
            import ThermoFisher.CommonCore.Data.Business as Business  # pylint: disable=redefined-outer-name
            import ThermoFisher.CommonCore.RawFileReader as _RawFileReader  # pylint: disable=redefined-outer-name
        except ImportError:
            continue
    return _test_dll_loaded()


def register_dll(search_paths=None):
    """Register the location of the Thermo RawFileReader DLL bundle with
    the Common Language Runtime interop system and load the .NET symbols
    used by this feature.

    Parameters
    ----------
    search_paths: list
        The paths to check along for the ThermoRawFileReader DLL bundle.

    """
    if search_paths is None:
        search_paths = []
    loaded = _register_dll(search_paths)
    if not loaded:
        msg = """The ThermoFisher.CommonCore libraries could not be located and loaded."""
        raise ImportError(msg)


def _test_dll_loaded():
    return _RawFileReader is not None


def _copy_double_array(src):
    """A quick and dirty implementation of the fourth technique shown in
    https://mail.python.org/pipermail/pythondotnet/2014-May/001525.html for
    copying a .NET Array[Double] to a NumPy ndarray[np.float64] via a raw
    memory copy.

    ``int_ptr_tp`` must be an integer type that can hold a pointer. On Python 2
    this is :class:`long`, and on Python 3 it is :class:`int`.
    """
    # When the input .NET array pointer is None, return an empty array. On Py2
    # this would happen automatically, but not on Py3, and perhaps not safely on
    # all Py2 because it relies on pythonnet and the .NET runtime properly checking
    # for nulls.
    if src is None:
        return np.array([], dtype=np.float64)
    dest = np.empty(len(src), dtype=np.float64)
    Marshal.Copy(
        src, 0,
        IntPtr.__overloads__[Int64](dest.__array_interface__['data'][0]),
        len(src))
    return dest


class RawReaderInterface(ScanDataSource):
    """:class:`~.ScanDataSource` implementation for Thermo's RawFileReader API.
    Not intended for direct instantiation.
    """

    def _scan_arrays(self, scan):
        scan_number = scan.scan_number + 1
        stats = self._source.GetScanStatsForScanNumber(scan_number)
        segscan = self._source.GetSegmentedScanFromScanNumber(scan_number, stats)
        mzs = _copy_double_array(segscan.Positions)
        inten = _copy_double_array(segscan.Intensities)
        return mzs, inten

    def _pick_peaks_vendor(self, scan, *args, **kwargs):
        scan_info = Business.Scan.FromFile(self._source, scan.scan_number + 1)
        if scan_info.HasCentroidStream:
            stream = self._source.GetCentroidStream(scan.scan_number + 1, 0)
            mzs = stream.Masses
            intens = stream.Intensities
            peaks = PeakSet([simple_peak(mzs[i], intens[i], 0.001) for i in range(len(mzs))])
            peaks.reindex()
            arrays = self._scan_arrays(scan)
            return PeakIndex(arrays[0], arrays[1], peaks)
        else:
            raise NotImplementedError()

    def _scan_id(self, scan):
        scan_number = scan.scan_number
        return _make_id(scan_number + 1)

    def _is_profile(self, scan):
        return not self._source.IsCentroidScanFromScanNumber(
            scan.scan_number + 1)

    def _polarity(self, scan):
        filter_string = self._filter_string(scan)
        return filter_string.data['polarity']

    def _scan_title(self, scan):
        return "%s %r" % (self._scan_id(scan), self._filter_string(scan))

    def _filter_string(self, scan):
        if scan.filter_string is None:
            scan_number = scan.scan_number
            scan.filter_string = FilterString(self._source.GetFilterForScanNumber(scan_number + 1).Filter)
        return scan.filter_string

    def _scan_index(self, scan):
        scan_number = scan.scan_number
        return scan_number

    def _scan_time(self, scan):
        scan_number = scan.scan_number
        return self._source.RetentionTimeFromScanNumber(scan_number + 1)

    def _ms_level(self, scan):
        scan_number = scan.scan_number
        f = self._source.GetFilterForScanNumber(scan_number + 1)
        return f.MSOrder

    def _isolation_window(self, scan):
        scan_number = scan.scan_number
        ms_level = self._ms_level(scan)
        width = 0
        trailer = self._trailer_values(scan)
        filt = self._source.GetFilterForScanNumber(scan_number + 1)
        seq_index = filt.MSOrder - 2
        try:
            # Fetch the isolation window width from the old location first, which
            # will be correct on old files, where the new API won't be right.
            width = trailer['MS%d Isolation Width' % ms_level]
        except KeyError:
            # Fall back to the new API, which is akin to our only hope here?
            width = filt.GetIsolationWidth(seq_index)
        width /= 2.0
        offset = filt.GetIsolationWidthOffset(seq_index)
        precursor_mz = filt.GetMass(seq_index)
        return IsolationWindow(width, precursor_mz + offset, width)

    def _trailer_values(self, scan):
        if scan.trailer_values is not None:
            return scan.trailer_values
        scan_number = scan.scan_number
        trailers = self._source.GetTrailerExtraInformation(scan_number + 1)
        scan.trailer_values = OrderedDict(
            zip([label.strip(":") for label in trailers.Labels], map(_try_number, trailers.Values)))
        return scan.trailer_values

    def _precursor_information(self, scan):
        scan_number = scan.scan_number
        filt = self._source.GetFilterForScanNumber(scan_number + 1)
        precursor_mz = filt.GetMass(filt.MSOrder - 2)
        trailers = self._trailer_values(scan)
        _precursor_mz = float(trailers.get("Monoisotopic M/Z", 0))
        if _precursor_mz > 0:
            precursor_mz = _precursor_mz

        # imitate proteowizard's firmware bug correction
        isolation_window = self._isolation_window(scan)
        if (isolation_window.upper + isolation_window.lower) / 2 <= 2.0:
            if (isolation_window.target - 3.0 > precursor_mz) or (isolation_window.target + 2.5 < precursor_mz):
                precursor_mz = isolation_window.target
        elif precursor_mz not in isolation_window:
            precursor_mz = isolation_window.target
        charge = int(trailers.get("Charge State", 0))
        if charge == 0:
            charge = ChargeNotProvided
        inten = 0
        precursor_scan_number = None
        precursor_scan_number = trailers.get('Master Scan Number')
        if precursor_scan_number == 0:
            precursor_scan_number = None
        if precursor_scan_number is not None:
            precursor_scan_number = int(precursor_scan_number) - 1
            if self.get_scan_by_index(precursor_scan_number).ms_level >= self._ms_level(scan):
                precursor_scan_number = None
        if precursor_scan_number is None:
            current_level = self._ms_level(scan)
            # We want to start looking for the most recent spectrum with the next lowest MS
            # level. The expecteation is that we couldn't determine the precursor scan accurately,
            # so we'll probe backwards until the next best candidate.
            current_index = self._scan_index(scan)
            # Use the index of the current scan to check the index of the previous scans at a lower
            # MS level. This may be None, in which case there is no earlier scan recorded.
            lookup = self._get_previous_scan_index_for_ms_level(current_index, current_level - 1)
            if lookup is not None:
                last_index = lookup
            else:
                last_index = current_index - 1
            i = 0
            while last_index >= 0 and i < 100:
                prev_scan = self.get_scan_by_index(last_index)
                if prev_scan.ms_level >= current_level:
                    last_index -= 1
                else:
                    precursor_scan_number = prev_scan._data.scan_number
                    break
                i += 1
        if precursor_scan_number is not None:
            precursor_scan_id = self.get_scan_by_index(precursor_scan_number).id
        else:
            import warnings
            warnings.warn("Could not resolve precursor scan for %s" % (self._scan_id(scan), ))
            precursor_scan_id = None
        return PrecursorInformation(
            precursor_mz, inten, charge, precursor_scan_id,
            source=self, product_scan_id=self._scan_id(scan))

    def _get_scan_segment(self, scan):
        trailer = self._trailer_values(scan)
        try:
            return int(trailer['Scan Segment'])
        except KeyError:
            return 1

    def _get_scan_event(self, scan):
        trailer = self._trailer_values(scan)
        try:
            return int(trailer['Scan Event'])
        except KeyError:
            return 1

    def _activation(self, scan):
        filter_string = self._filter_string(scan)
        tandem_sequence = filter_string.get("tandem_sequence")
        # If the tandem sequence exists, the last entry is the most recent tandem acquisition.
        # It will list contain one or more activation types. Alternatively, multiple activations
        # of the same precursor may exist in the list as separate events in the tandem sequence.
        if tandem_sequence is not None:
            activation_event = tandem_sequence[-1]
            activation_type = list(activation_event.get("activation_type"))
            has_supplemental_activation = filter_string.get("supplemental_activation")

            if activation_type is not None:
                energy = list(activation_event.get("activation_energy"))
                if len(tandem_sequence) > 1:
                    prev_event = tandem_sequence[-2]
                    # Merge previous tandem sequences of the same precursor
                    if abs(prev_event['isolation_mz'] - activation_event['isolation_mz']) < 1e-3:
                        activation_type = list(prev_event.get("activation_type")) + activation_type
                        energy = list(prev_event.get("activation_energy")) + energy
                        has_supplemental_activation = True

                if has_supplemental_activation and len(activation_type) > 1:
                    activation_type.append(supplemental_term_map[
                        dissociation_methods_map[activation_type[-1]]])
                if len(activation_type) == 1:
                    return ActivationInformation(activation_type[0], energy[0])
                else:
                    return MultipleActivationInformation(activation_type, energy)
        return None

    def _acquisition_information(self, scan):
        fline = self._filter_string(scan)
        event = self._get_scan_event(scan)
        trailer_extras = self._trailer_values(scan)
        traits = {
            'preset scan configuration': event,
            'filter string': fline,
        }
        cv = fline.get("compensation_voltage")
        if cv is not None:
            traits[FAIMS_compensation_voltage] = cv
        event = ScanEventInformation(
            self._scan_time(scan),
            injection_time=unitfloat(trailer_extras.get('Ion Injection Time (ms)', 0.0), 'millisecond'),
            window_list=[ScanWindow(
                fline.get("scan_window")[0], fline.get("scan_window")[1])], traits=traits)
        return ScanAcquisitionInformation("no combination", [event])

    def _instrument_configuration(self, scan):
        fline = self._filter_string(scan)
        try:
            confid = self._analyzer_to_configuration_index[analyzer_map[fline.data.get("analyzer")]]
            return self._instrument_config[confid]
        except KeyError:
            return None

    def _annotations(self, scan):
        fline = self._filter_string(scan)
        trailer_extras = self._trailer_values(scan)
        annots = {
            "filter string": fline,
        }
        microscans = trailer_extras.get("Micro Scan Count")
        if microscans is not None:
            annots['[Thermo Trailer Extra]Micro Scan Count'] = float(microscans)
        scan_segment = trailer_extras.get("Scan Segment")
        if scan_segment is not None:
            annots['[Thermo Trailer Extra]Scan Segment'] = int(scan_segment)
        scan_event = trailer_extras.get("Scan Event")
        if scan_event is not None:
            annots['[Thermo Trailer Extra]Scan Event'] = int(scan_event)
        mono_mz = float(trailer_extras.get("Monoisotopic M/Z", 0))
        if mono_mz is not None and mono_mz > 0:
            annots['[Thermo Trailer Extra]Monoisotopic M/Z'] = mono_mz
        hcd_ev = _trailer_float(trailer_extras.get('HCD Energy eV'))
        if hcd_ev is not None and hcd_ev > 0:
            annots['[Thermo Trailer Extra]HCD Energy eV'] = float(hcd_ev)
        hcd_energies = trailer_extras.get('HCD Energy')
        if hcd_energies is not None:
            if isinstance(hcd_energies, basestring) and not hcd_energies.strip():
                pass
            else:
                annots['[Thermo Trailer Extra]HCD Energy'] = hcd_energies
        return annots


def _trailer_float(value):
    if value is None:
        return None
    try:
        value = value.strip()
    except AttributeError:
        return value
    if not value:
        return None
    try:
        return float(value)
    except ValueError:
        return None


class ThermoRawLoader(RawReaderInterface, RandomAccessScanSource, _RawFileMetadataLoader):
    """Reads scans from Thermo Fisher RAW files directly. Provides both iterative and
    random access.
    """

    def __init__(self, source_file, _load_metadata=True, **kwargs):
        if not _test_dll_loaded():
            register_dll()
            if not _test_dll_loaded():
                raise ValueError(
                    "Could not load .NET runtime or library. Verify `pythonnet` "
                    "is installed correctly and the RawFileReader library is registered")
        self.source_file = source_file
        self._source_impl = None
        # self._source = _RawFileReader.RawFileReaderAdapter.FileFactory(source_file)
        # self._source.SelectInstrument(Business.Device.MS, 1)
        self._producer = None
        self._scan_type_index = dict()

        self._method = None
        self._analyzer_to_configuration_index = {}
        self._instrument_config = {}

        self.make_iterator()
        self.initialize_scan_cache()
        self._first_scan_time = self.get_scan_by_index(0).scan_time
        self._last_scan_time = self.get_scan_by_index(
            self._source.RunHeaderEx.LastSpectrum - 1).scan_time
        self._index = self._pack_index()

        if _load_metadata:
            self._method = self._parse_method()
            self._build_scan_type_index()
            self._get_instrument_info()

    def software_list(self) -> List[software.Software]:
        inst_data = self._source.GetInstrumentData()
        sw_list = [
            software.Software('Xcalibur', 'Xcalibur', inst_data.SoftwareVersion)
        ]
        return sw_list

    @property
    def _source(self):
        if self._source_impl is None:
            self._source_impl = _open_raw_file(self.source_file)
            self._source_impl.SelectInstrument(Business.Device.MS, 1)
        return self._source_impl

    @_source.setter
    def _source(self, value):
        self._source_impl = value

    def _has_ms1_scans(self):
        if self._scan_type_index:
            return 1 in self._scan_type_index
        else:
            # metadata has not been loaded so best to assume there is
            return True

    def _has_msn_scans(self):
        if self._scan_type_index:
            return max(self._scan_type_index) > 1
        else:
            # metadata has not been loaded so best to assume there is
            return True

    def has_msn_scans(self):
        return self._has_msn_scans()

    def has_ms1_scans(self):
        return self._has_ms1_scans()

    def __reduce__(self):
        reduced = self.__class__, (self.source_file, False), self.__getstate__()
        self._close_handle()
        return reduced

    def __getstate__(self):
        state = {
            "method": self._method,
            "scan_type_index": self._scan_type_index,
            "analyzer_to_configuration_index": self._analyzer_to_configuration_index,
            "instrument_config": self._instrument_config,
            "previous_ms_levels": self._previous_ms_levels,
        }
        return state

    def __setstate__(self, state):
        self._method = state['method']
        self._scan_type_index = state['scan_type_index']
        self._analyzer_to_configuration_index = state['analyzer_to_configuration_index']
        self._instrument_config = state['instrument_config']
        self._previous_ms_levels = state['previous_ms_levels']

    @property
    def index(self):
        """Accesses the scan index

        Returns
        -------
        :class:`collections.OrderedDict`
            Maps scan ID to index
        """
        return self._index

    def __len__(self):
        return len(self.index)

    def __repr__(self):
        return "ThermoRawLoader(%r)" % (self.source_file)

    def _close_handle(self):
        if self._source is not None:
            self._source.Close()
            self._source = None

    def close(self):
        """Close the underlying file reader.
        """
        self._close_handle()
        self._dispose()

    def __del__(self):
        self.close()

    def reset(self):
        self.make_iterator(None)
        self.initialize_scan_cache()

    def _pack_index(self):
        index = OrderedDict()
        for sn in range(self._source.RunHeaderEx.FirstSpectrum,
                        self._source.RunHeaderEx.LastSpectrum + 1):
            index[_make_id(sn)] = sn
        return index

    def _get_instrument_model_name(self):
        return self._source.GetInstrumentData().Model

    def _get_instrument_serial_number(self):
        return self._source.GetInstrumentData().SerialNumber

    def _parse_method(self):
        try:
            method_count = self._source.InstrumentMethodsCount
            if method_count == 0:
                return _InstrumentMethod('')
            # the data acquisition method should be the last method
            return _InstrumentMethod(self._source.GetInstrumentMethod(method_count - 1))
        except NullReferenceException:   # pylint: disable=broad-except
            return _InstrumentMethod('')

    def _scan_time_to_scan_number(self, scan_time):
        scan_number = self._source.ScanNumberFromRetentionTime(scan_time) - 1
        return scan_number

    def samples(self):
        """Describe the sample(s) used to generate the mass spectrometry
        data contained in this file.

        Returns
        -------
        :class:`list` of :class:`~.Sample`
        """
        result = []
        si = self._source.SampleInformation
        raw_name = os.path.basename(self.source_file_name)
        sample = Sample(si.SampleId or raw_name)
        sample.name = si.SampleName or si.SampleId
        if si.SampleVolume:
            sample.parameters['sample volume'] = si.SampleVolume
        if si.SampleWeight:
            sample.parameters['sample mass'] = si.SampleWeight
        if si.Vial:
            sample.parameters['sample vial'] = si.Vial
        if si.Barcode:
            sample.parameters['sample barcode'] = si.Barcode
        result.append(sample)
        return result

    def get_scan_by_index(self, index):
        """Retrieve the scan object for the specified scan index.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        index: int
            The index to get the scan for

        Returns
        -------
        Scan
        """
        scan_number = int(index)
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = ThermoRawScanPtr(scan_number)
            if not package.validate(self):
                raise IndexError(index)
            scan = Scan(package, self)
            self._scan_cache[scan_number] = scan
            return scan

    def get_scan_by_id(self, scan_id):
        """Retrieve the scan object for the specified scan id.

        If the scan object is still bound and in memory somewhere,
        a reference to that same object will be returned. Otherwise,
        a new object will be created.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        Scan
        """
        scan_number = int(str(scan_id).replace(_id_template, '')) - 1
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = ThermoRawScanPtr(scan_number)
            if not package.validate(self):
                raise KeyError(str(scan_id))
            scan = Scan(package, self)
            self._scan_cache[scan_number] = scan
            return scan

    def get_scan_by_time(self, time):
        """Retrieve the scan object for the specified scan time.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        time : float
            The time to get the nearest scan from

        Returns
        -------
        Scan
        """
        if time < self._first_scan_time:
            time = self._first_scan_time
        elif time > self._last_scan_time:
            time = self._last_scan_time
        scan_number = self._scan_time_to_scan_number(time)
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = ThermoRawScanPtr(scan_number)
            scan = Scan(package, self)
            self._scan_cache[scan_number] = scan
            return scan

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True, **kwargs):
        """Reconstruct an iterator which will start from the scan matching one of ``scan_id``,
        ``rt``, or ``index``. Only one may be provided.

        After invoking this method, the iterator this object wraps will be changed to begin
        yielding scan bunchs (or single scans if ``grouped`` is ``False``).

        Arguments
        ---------
        scan_id: str, optional
            Start from the scan with the specified id.
        rt: float, optional
            Start from the scan nearest to specified time (in minutes) in the run. If no
            exact match is found, the nearest scan time will be found, rounded up.
        index: int, optional
            Start from the scan with the specified index.
        require_ms1: bool, optional
            Whether the iterator must start from an MS1 scan. True by default.
        grouped: bool, optional
            whether the iterator should yield scan bunches or single scans. True by default.
        """
        if scan_id is not None:
            scan_number = int(str(scan_id).replace(_id_template, '')) - 1
        elif index is not None:
            scan_number = int(index)
        elif rt is not None:
            scan_number = self._scan_time_to_scan_number(rt)
        if require_ms1:
            start_index = scan_number
            while start_index != 0:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            scan_number = start_index
        iterator = self._make_pointer_iterator(start_index=scan_number)
        if grouped:
            self._producer = self._scan_group_iterator(iterator, grouped, **kwargs)
        else:
            self._producer = self._single_scan_iterator(
                iterator, grouped, **kwargs)
        return self

    def _make_scan_index_producer(self, start_index=None, start_time=None):
        if start_index is not None:
            return range(start_index, self._source.RunHeaderEx.LastSpectrum)
        elif start_time is not None:
            start_index = self._scan_time_to_scan_number(start_time)
            while start_index != 0:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            return range(start_index, self._source.RunHeaderEx.LastSpectrum)
        else:
            return range(0, self._source.RunHeaderEx.LastSpectrum)

    def _make_pointer_iterator(self, start_index=None, start_time=None):
        iterator = self._make_scan_index_producer(start_index, start_time)
        for i in iterator:
            yield ThermoRawScanPtr(i)

    def _make_default_iterator(self):
        return self._make_pointer_iterator()

    def _make_cache_key(self, scan):
        return scan._data.scan_number

    def next(self):
        return next(self._producer)
